import os
import pprint
import json
import sys
import re
import urllib.parse
import requests
import inspect
import importlib
import textwrap
import pip  # noqa: F401
import subprocess
import copy
from string import Template
from yapf.yapflib import yapf_api

from biothings.utils.hub_db import get_data_plugin, get_src_dump, get_src_master
from biothings.utils.common import rmdashfr, get_class_from_classpath
from biothings.utils.loggers import get_logger
from biothings import config as btconfig

from biothings.utils.manager import BaseSourceManager
from biothings.hub.dataload.dumper import LastModifiedHTTPDumper, LastModifiedFTPDumper  # noqa: F402
from biothings.hub.dataload.uploader import BaseSourceUploader, ParallelizedSourceUploader  # noqa: F401
from biothings.hub.dataload.storage import IgnoreDuplicatedStorage, BasicStorage, MergerStorage  # noqa: F401
from biothings.hub.dataplugin.manager import GitDataPlugin, ManualDataPlugin


class AssistantException(Exception):
    pass

class LoaderException(Exception):
    pass


class BasePluginLoader(object):

    loader_type = None  # set in subclass

    def __init__(self, plugin_name):
        self.plugin_name = plugin_name
        self.setup_log()

    def setup_log(self):
        """Setup and return a logger instance"""
        self.logger, self.logfile = get_logger('loader_%s' % self.plugin_name)

    def get_plugin_obj(self):
        dp = get_data_plugin()
        plugin = dp.find_one({"_id": self.plugin_name})
        if not plugin.get("download", {}).get("data_folder"):
            raise LoaderException("Can't find data_folder, not available yet ?")
        return plugin

    def invalidate_plugin(self, error):
        self.logger.exception("Invalidate plugin '%s' because: %s" %
                              (self.plugin_name, error))
        # flag all plugin associated (there should only one though, but no need to care here)
        for klass in self.__class__.data_plugin_manager[self.plugin_name]:
            klass.data_plugin_error = error
        raise LoaderException(error)

    def can_load_plugin(self):
        """
        Return True if loader is able to load plugin (check data folder content)
        """
        raise NotImplementedError("implement 'can_load_plugin' in subclass")

    def load_plugin(self):
        """
        Load plugin and register its components
        """
        raise NotImplementedError("implement 'load_plugin' in subclass")


class ManifestBasedPluginLoader(BasePluginLoader):

    loader_type = "manifest"

    # should match a _dict_for_***
    dumper_registry = {
        "http": LastModifiedHTTPDumper,
        "https": LastModifiedHTTPDumper,
        "ftp": LastModifiedFTPDumper
    }

    def _dict_for_base(self, data_url):
        if type(data_url) == str:
            data_url = [data_url]
        return {
            "SRC_NAME": self.plugin_name,
            "SRC_ROOT_FOLDER": os.path.join(btconfig.DATA_ARCHIVE_ROOT, self.plugin_name),
            "SRC_URLS": data_url
        }

    def _dict_for_http(self, data_url):
        return self._dict_for_base(data_url)

    def _dict_for_https(self, data_url):
        d = self._dict_for_http(data_url)
        # not secure, but we want to make sure things will work as much as possible...
        d["VERIFY_CERT"] = False
        return d

    def _dict_for_ftp(self, data_url):
        return self._dict_for_base(data_url)

    def can_load_plugin(self):
        plugin = self.get_plugin_obj()
        df = plugin["download"]["data_folder"]
        if "manifest.json" in os.listdir(df) and os.path.exists(os.path.join(df,"manifest.json")):
            return True
        else:
            return False

    def load_plugin(self):
        plugin = self.get_plugin_obj()
        df = plugin["download"]["data_folder"]
        if os.path.exists(df):
            mf = os.path.join(df, "manifest.json")
            if os.path.exists(mf):
                try:
                    self.logger.debug("Loading manifest: %s" % mf)
                    self.interpret_manifest(mf)
                except Exception as e:
                    self.invalidate_plugin("Error loading manifest: %s" %
                                           str(e))
            else:
                self.logger.info("No manifest found for plugin: %s" %
                                 plugin["plugin"]["url"])
                self.invalidate_plugin("No manifest found")
        else:
            self.invalidate_plugin("Missing plugin folder '%s'" % df)

    def get_code_for_mod_name(self, mod_name):
        try:
            mod, funcname = map(str.strip, mod_name.split(":"))
        except ValueError as e:
            raise AssistantException("'Wrong format for '%s', it must be defined following format 'module:func': %s" % (mod_name, e))
        modpath = self.plugin_name + "." + mod
        pymod = importlib.import_module(modpath)
        # reload in case we need to refresh plugin's code
        importlib.reload(pymod)
        assert funcname in dir(pymod), "%s not found in module %s" % (funcname, pymod)
        func = getattr(pymod, funcname)
        # fetch source and indent to class method level in the template
        strfunc = inspect.getsource(func)
        # always indent with spaces, normalize to avoid mixed indenting chars
        indentfunc = textwrap.indent(strfunc.replace("\t", "    "), prefix="    ")

        return indentfunc, funcname

    def get_dumper_dynamic_class(self, dumper_section, metadata):
        if dumper_section.get("data_url"):
            if not type(dumper_section["data_url"]) is list:
                durls = [dumper_section["data_url"]]
            else:
                durls = dumper_section["data_url"]
            schemes = set(
                [urllib.parse.urlsplit(durl).scheme for durl in durls])
            # https = http regarding dumper generation
            if len(set([sch.replace("https", "http") for sch in schemes])) > 1:
                raise AssistantException("Manifest specifies URLs of different types (%s), " % schemes
                                         + "expecting only one")
            scheme = schemes.pop()
            klass = dumper_section.get("class")
            confdict = getattr(self, "_dict_for_%s" % scheme)(durls)
            dumper_class = None
            if klass:
                dumper_class = get_class_from_classpath(klass)
                confdict["BASE_CLASSES"] = klass
            else:
                dumper_class = self.dumper_registry.get(scheme)
                confdict[
                    "BASE_CLASSES"] = "biothings.hub.dataload.dumper.%s" % dumper_class.__name__
            if not dumper_class:
                raise AssistantException(
                    "No dumper class registered to handle scheme '%s'" %
                    scheme)
            if metadata:
                confdict["__metadata__"] = metadata
            else:
                confdict["__metadata__"] = {}

            if dumper_section.get("release"):
                indentfunc, func = self.get_code_for_mod_name(dumper_section["release"])
                assert func != "set_release", "'set_release' is a reserved method name, pick another name"
                confdict["SET_RELEASE_FUNC"] = """
%s

    def set_release(self):
        self.release = self.%s()
""" % (indentfunc, func)

            else:
                confdict["SET_RELEASE_FUNC"] = ""

            dklass = None
            pnregex = r"^[A-z_][\w\d]+$"
            assert re.compile(pnregex).match(self.plugin_name), \
                "Incorrect plugin name '%s' (doesn't match regex '%s'" % (self.plugin_name, pnregex)
            dumper_name = self.plugin_name.capitalize() + "Dumper"
            '%s'
            try:
                if hasattr(btconfig, "DUMPER_TEMPLATE"):
                    tpl_file = btconfig.DUMPER_TEMPLATE
                else:
                    # default: assuming in ..../biothings/hub/dataplugin/
                    curmodpath = os.path.realpath(__file__)
                    tpl_file = os.path.join(os.path.dirname(curmodpath),
                                            "dumper.py.tpl")
                tpl = Template(open(tpl_file).read())
                confdict["DUMPER_NAME"] = dumper_name
                confdict["SRC_NAME"] = self.plugin_name
                if dumper_section.get("schedule"):
                    schedule = """'%s'""" % dumper_section["schedule"]
                else:
                    schedule = "None"
                confdict["SCHEDULE"] = schedule
                confdict["UNCOMPRESS"] = dumper_section.get(
                    "uncompress") or False
                pystr = tpl.substitute(confdict)
                #print(pystr)
                import imp
                code = compile(pystr, "<string>", "exec")
                mod = imp.new_module(self.plugin_name)
                exec(code, mod.__dict__, mod.__dict__)
                dklass = getattr(mod, dumper_name)
                # we need to inherit from a class here in this file so it can be pickled
                assisted_dumper_class = type(
                    "AssistedDumper_%s" % self.plugin_name, (
                        AssistedDumper,
                        dklass,
                    ), {})
                assisted_dumper_class.python_code = pystr

                return assisted_dumper_class

            except Exception:
                self.logger.exception("Can't generate dumper code for '%s'" % self.plugin_name)
                raise
        else:
            raise AssistantException(
                "Invalid manifest, expecting 'data_url' key in 'dumper' section"
            )

    def get_uploader_dynamic_class(self, uploader_section, metadata):
        if uploader_section.get("parser"):
            uploader_name = self.plugin_name.capitalize() + "Uploader"
            confdict = {
                "SRC_NAME": self.plugin_name,
                "UPLOADER_NAME": uploader_name
            }
            try:
                mod, func = uploader_section.get("parser").split(":")
                confdict["PARSER_MOD"] = mod
                confdict["PARSER_FUNC"] = func
            except ValueError:
                raise AssistantException("'parser' must be defined as 'module:parser_func' but got: '%s'" %
                                         uploader_section["parser"])
            try:
                ondups = uploader_section.get("on_duplicates")
                if ondups and ondups != "error":
                    if ondups == "merge":
                        storage_class = "biothings.hub.dataload.storage.MergerStorage"
                    elif ondups == "ignore":
                        storage_class = "biothings.hub.dataload.storage.IgnoreDuplicatedStorage"
                else:
                    storage_class = "biothings.hub.dataload.storage.BasicStorage"
                if uploader_section.get("ignore_duplicates"):
                    raise AssistantException(
                        "'ignore_duplicates' key not supported anymore, "
                        + "use 'on_duplicates' : 'error|ignore|merge'")
                confdict["STORAGE_CLASS"] = storage_class
                # default is not ID conversion at all
                confdict["IMPORT_IDCONVERTER_FUNC"] = ""
                confdict["IDCONVERTER_FUNC"] = None
                confdict["CALL_PARSER_FUNC"] = "parser_func(data_folder)"
                if uploader_section.get("keylookup"):
                    assert self.__class__.keylookup, "Plugin %s needs _id conversion " % self.plugin_name + \
                                                     "but no keylookup instance was found"
                    self.logger.info("Keylookup conversion required: %s" %
                                     uploader_section["keylookup"])
                    klmod = inspect.getmodule(self.__class__.keylookup)
                    confdict[
                        "IMPORT_IDCONVERTER_FUNC"] = "from %s import %s" % (
                            klmod.__name__, self.__class__.keylookup.__name__)
                    convargs = ",".join([
                        "%s=%s" % (k, v)
                        for k, v in uploader_section["keylookup"].items()
                    ])
                    confdict["IDCONVERTER_FUNC"] = "%s(%s)" % (
                        self.__class__.keylookup.__name__, convargs)
                    confdict[
                        "CALL_PARSER_FUNC"] = "self.__class__.idconverter(parser_func)(data_folder)"
                if metadata:
                    confdict["__metadata__"] = metadata
                else:
                    confdict["__metadata__"] = {}

                if hasattr(btconfig, "DUMPER_TEMPLATE"):
                    tpl_file = btconfig.DUMPER_TEMPLATE
                else:
                    # default: assuming in ..../biothings/hub/dataplugin/
                    curmodpath = os.path.realpath(__file__)
                    tpl_file = os.path.join(os.path.dirname(curmodpath),
                                            "uploader.py.tpl")
                tpl = Template(open(tpl_file).read())

                if uploader_section.get("parallelizer"):
                    indentfunc, func = self.get_code_for_mod_name(uploader_section["parallelizer"])
                    assert func != "jobs", "'jobs' is a reserved method name, pick another name"
                    confdict[
                        "BASE_CLASSES"] = "biothings.hub.dataload.uploader.ParallelizedSourceUploader"
                    confdict["IMPORT_FROM_PARALLELIZER"] = ""
                    confdict["JOBS_FUNC"] = """
%s
    def jobs(self):
        return self.%s()
""" % (indentfunc, func)
                else:
                    confdict[
                        "BASE_CLASSES"] = "biothings.hub.dataload.uploader.BaseSourceUploader"
                    confdict["JOBS_FUNC"] = ""

                if uploader_section.get("mapping"):
                    indentfunc, func = self.get_code_for_mod_name(uploader_section["mapping"])
                    assert func != "get_mapping", "'get_mapping' is a reserved class method name, pick another name"
                    confdict["MAPPING_FUNC"] = """
    @classmethod
%s

    @classmethod
    def get_mapping(cls):
        return cls.%s()
""" % (indentfunc, func)
                else:
                    confdict["MAPPING_FUNC"] = ""

                pystr = tpl.substitute(confdict)
                #print(pystr)
                import imp
                code = compile(pystr, "<string>", "exec")
                mod = imp.new_module(self.plugin_name)
                exec(code, mod.__dict__, mod.__dict__)
                uklass = getattr(mod, uploader_name)
                # we need to inherit from a class here in this file so it can be pickled
                assisted_uploader_class = type(
                    "AssistedUploader_%s" % self.plugin_name, (
                        AssistedUploader,
                        uklass,
                    ), {})
                assisted_uploader_class.python_code = pystr

                return assisted_uploader_class

            except Exception as e:
                self.logger.exception("Error loading plugin: %s" % e)
                raise AssistantException("Can't interpret manifest: %s" % e)
        else:
            raise AssistantException(
                "Invalid manifest, expecting 'parser' key in 'uploader' section"
            )

    def interpret_manifest(self, manifest_file):
        manifest = json.load(open(manifest_file))
        # start with requirements before importing anything
        if manifest.get("requires"):
            reqs = manifest["requires"]
            if not type(reqs) == list:
                reqs = [reqs]
            for req in reqs:
                self.logger.info("Install requirement '%s'" % req)
                subprocess.check_call(
                    [sys.executable, '-m', 'pip', 'install', req])
        if manifest.get("dumper"):
            assisted_dumper_class = self.get_dumper_dynamic_class(
                manifest["dumper"], manifest.get("__metadata__"))
            assisted_dumper_class.DATA_PLUGIN_FOLDER = os.path.dirname(
                manifest_file)
            self.__class__.dumper_manager.register_classes(
                [assisted_dumper_class])
            # register class in module so it can be pickled easily
            sys.modules["biothings.hub.dataplugin.assistant"].__dict__[
                "AssistedDumper_%s" % self.plugin_name] = assisted_dumper_class

        if manifest.get("uploader"):
            assisted_uploader_class = self.get_uploader_dynamic_class(
                manifest["uploader"], manifest.get("__metadata__"))
            assisted_uploader_class.DATA_PLUGIN_FOLDER = os.path.dirname(
                manifest_file)
            self.__class__.uploader_manager.register_classes(
                [assisted_uploader_class])
            # register class in module so it can be pickled easily
            sys.modules["biothings.hub.dataplugin.assistant"].__dict__[
                "AssistedUploader_%s" %
                self.plugin_name] = assisted_uploader_class


class AdvancedPluginLoader(BasePluginLoader):

    loader_type = "advanced"

    def can_load_plugin(self):
        plugin = self.get_plugin_obj()
        df = plugin["download"]["data_folder"]
        if "__init__.py" in os.listdir(df):
            return True
        else:
            return False

    def load_plugin(self):
        plugin = self.get_plugin_obj()
        df = plugin["download"]["data_folder"]
        if os.path.exists(df):
            # we assume there's a __init__ module exposing Dumper and Uploader classes
            # as necessary
            modpath = df.split("/")[-1]
            # before registering, process optional requirements.txt
            reqfile = os.path.join(df,"requirements.txt")
            if os.path.exists(reqfile):
                self.logger.info("Installing requirements from %s for plugin '%s'" % (reqfile,self.plugin_name))
                subprocess.check_call([sys.executable,"-m","pip","install","-r",reqfile])
            # submit to managers to register datasources
            self.logger.info("Registering '%s' to dump/upload managers" % modpath)
            # register dumpers if any
            try:
                self.__class__.dumper_manager.register_source(modpath)
            except Exception as e:
                self.logger.info("Couldn't register dumper from module '%s': %s" % (modpath,e))
            # register uploaders if any
            try:
                self.__class__.uploader_manager.register_source(modpath)
            except Exception as e:
                self.logger.info("Couldn't register uploader from module '%s': %s" % (modpath,e))
        else:
            self.invalidate_plugin("Missing plugin folder '%s'" % df)


class BaseAssistant(object):

    plugin_type = None  # to be defined in subblass
    data_plugin_manager = None  # set by assistant manager
    dumper_manager = None  # set by assistant manager
    uploader_manager = None  # set by assistant manager
    keylookup = None  # set by assistant manager

    # known plugin loaders
    loaders = {
        "manifest": ManifestBasedPluginLoader,
        "advanced": AdvancedPluginLoader,
    }

    def __init__(self, url):
        self.url = url
        self._plugin_name = None
        self._src_folder = None
        self._loader = None
        self.logfile = None
        self.setup_log()

    def setup_log(self):
        """Setup and return a logger instance"""
        self.logger, self.logfile = get_logger('assistant_%s' %
                                               self.__class__.plugin_type)

    def register_loader(self):
        dp = get_data_plugin()
        dp.update({"_id": self.plugin_name}, {
            "$set": { "plugin.loader": self.loader.loader_type }
            },
            upsert=True)

    @property
    def loader(self):
        """
        Return loader object able to interpret plugin's folder content
        """
        if not self._loader:
            # iterate over known loaders, the first one which can interpret plugin content is kept
            for klass in self.loaders.values():
                # propagate managers
                klass.dumper_manager = self.dumper_manager
                klass.uploader_manager = self.uploader_manager
                klass.data_plugin_manager = self.data_plugin_manager
                klass.keylookup = self.keylookup
                loader = klass(self.plugin_name)
                if loader.can_load_plugin():
                    self._loader = loader
                    self.logger.info("For plugin '%s', selecting loader %s" % (self.plugin_name, self._loader))
                    self.register_loader()
                    break
                else:
                    self.logger.debug("Loader %s can't load plugin '%s'" % (loader, self.plugin_name))
                    continue
        return self._loader

    @property
    def plugin_name(self):
        """
        Return plugin name, parsed from self.url and set self._src_folder as
        path to folder containing dataplugin source code
        """
        raise NotImplementedError("implement 'plugin_name' in subclass")

    def handle(self):
        """Access self.url and do whatever is necessary to bring code to life within the hub...
        (hint: that may involve creating a dumper on-the-fly and register that dumper to
        a manager...)
        """
        raise NotImplementedError("implement 'handle' in subclass")

    def can_handle(self):
        """Return true if assistant can handle the code"""
        raise NotImplementedError("implement 'can_handle' in subclass")

    def load_plugin(self):
        """
        Load plugin and register its components
        """
        raise NotImplementedError("implement 'load_plugin' in subclass")



class AssistedDumper(object):
    DATA_PLUGIN_FOLDER = None


class AssistedUploader(object):
    DATA_PLUGIN_FOLDER = None


class GithubAssistant(BaseAssistant):

    plugin_type = "github"

    @property
    def plugin_name(self):
        if not self._plugin_name:
            split = urllib.parse.urlsplit(self.url)
            self._plugin_name = os.path.basename(split.path).replace(".git", "")
            self._src_folder = os.path.join(btconfig.DATA_PLUGIN_FOLDER, self.plugin_name)
        return self._plugin_name

    def can_handle(self):
        # analyze headers to guess type of required assitant
        try:
            headers = requests.head(self.url).headers
            if headers.get("server").lower() == "github.com":
                return True
        except Exception as e:
            self.logger.error("%s plugin can't handle URL '%s': %s" %
                             (self.plugin_type, self.url, e))
            return False

    def get_classdef(self):
        # generate class dynamically and register
        confdict = {
            "SRC_NAME": self.plugin_name,
            "GIT_REPO_URL": self.url,
            "SRC_ROOT_FOLDER": self._src_folder
        }
        # TODO: store confdict in hubconf collection
        k = type("AssistedGitDataPlugin_%s" % self.plugin_name,
                 (GitDataPlugin, ), confdict)
        return k

    def handle(self):
        assert self.__class__.data_plugin_manager, "Please set data_plugin_manager attribute"
        klass = self.get_classdef()
        self.__class__.data_plugin_manager.register_classes([klass])


class LocalAssistant(BaseAssistant):

    plugin_type = "local"

    @property
    def plugin_name(self):
        if not self._plugin_name:
            split = urllib.parse.urlsplit(self.url)
            # format local://pluginname so it's in hostname.
            # if path is set, it means format is  local://subdir/pluginname
            # and we don't support that for import reason (we would need to
            # add .../plugins/subdir to sys.path, not impossible but might have side effects
            # so for now we stay on the safe (and also let's remember 1st version of
            # MS DOS didn't support subdirs, so I guess we're on the right path :))
            assert not split.path, "It seems URL '%s' references a sub-directory (%s)," % (self.url, split.hostname) \
                                   + " with plugin name '%s', sub-directories are not supported (yet)" % split.path.strip("/")
            # don't use hostname here because it's lowercased, netloc isn't
            # (and we're matching directory names on the filesystem, it's case-sensitive)
            self._plugin_name = os.path.basename(split.netloc)
            self._src_folder = os.path.join(btconfig.DATA_PLUGIN_FOLDER, self._plugin_name)
        return self._plugin_name

    def can_handle(self):
        if self.url.startswith(self.__class__.plugin_type + "://"):
            return True
        else:
            return False

    def get_classdef(self):
        # generate class dynamically and register
        confdict = {
            "SRC_NAME": self.plugin_name,
            "SRC_ROOT_FOLDER": self._src_folder
        }
        k = type("AssistedManualDataPlugin_%s" % self.plugin_name,
                 (ManualDataPlugin, ), confdict)
        return k

    def handle(self):
        assert self.__class__.data_plugin_manager, "Please set data_plugin_manager attribute"
        klass = self.get_classdef()
        self.__class__.data_plugin_manager.register_classes([klass])


class AssistantManager(BaseSourceManager):

    def __init__(self,
                 data_plugin_manager,
                 dumper_manager,
                 uploader_manager,
                 keylookup=None,
                 default_export_folder="hub/dataload/sources",
                 *args,
                 **kwargs):
        super(AssistantManager, self).__init__(*args, **kwargs)
        self.data_plugin_manager = data_plugin_manager
        self.dumper_manager = dumper_manager
        self.uploader_manager = uploader_manager
        self.keylookup = keylookup
        if not os.path.exists(btconfig.DATA_PLUGIN_FOLDER):
            os.makedirs(btconfig.DATA_PLUGIN_FOLDER)
        self.default_export_folder = default_export_folder
        # register data plugin folder in python path so we can import
        # plugins (sub-folders) as packages
        sys.path.insert(0, btconfig.DATA_PLUGIN_FOLDER)
        self.logfile = None
        self.setup_log()

    def setup_log(self):
        """Setup and return a logger instance"""
        self.logger, self.logfile = get_logger('assistantmanager')

    def create_instance(self, klass, url):
        return klass(url)

    def configure(self, klasses=[GithubAssistant, LocalAssistant, ]):
        self.register_classes(klasses)

    def register_classes(self, klasses):
        for klass in klasses:
            klass.data_plugin_manager = self.data_plugin_manager
            klass.dumper_manager = self.dumper_manager
            klass.uploader_manager = self.uploader_manager
            klass.keylookup = self.keylookup
            self.register[klass.plugin_type] = klass

    def submit(self, url):
        # submit url to all registered assistants (in order)
        # and return the first claiming it can handle that URLs
        for typ in self.register:
            aklass = self.register[typ]
            inst = self.create_instance(aklass, url)
            if inst.can_handle():
                return inst
        return None

    def unregister_url(self, url=None, name=None):
        dp = get_data_plugin()
        if url:
            url = url.strip()
            doc = dp.find_one({"plugin.url": url})
        elif name:
            doc = dp.find_one({"_id": name})
            url = doc["plugin"]["url"]
        else:
            raise ValueError("Specify 'url' or 'name'")
        if not doc:
            raise AssistantException(
                "Plugin is not registered (url=%s, name=%s)" % (url, name))
        # should be only one but just in case
        dp.remove({"_id": doc["_id"]})
        # delete plugin code so it won't be auto-register
        # by 'local' plugin assistant (issue studio #7)
        if doc.get("download", {}).get("data_folder"):
            codefolder = doc["download"]["data_folder"]
            self.logger.info("Delete plugin source code in '%s'" % codefolder)
            rmdashfr(codefolder)
        assistant = self.submit(url)
        try:
            self.data_plugin_manager.register.pop(assistant.plugin_name)
        except KeyError:
            raise AssistantException("Plugin '%s' is not registered" % url)
        self.dumper_manager.register.pop(assistant.plugin_name, None)
        self.uploader_manager.register.pop(assistant.plugin_name, None)

    def register_url(self, url):
        url = url.strip()
        dp = get_data_plugin()
        if dp.find_one({"plugin.url": url}):
            self.logger.info("Plugin '%s' already registered" % url)
            return
        assistant = self.submit(url)
        self.logger.info("For data-plugin URL '%s', selected assistant is: %s" % (url, assistant))
        if assistant:
            # register plugin info
            # if a github url was used, by default, we assume it's a manifest-based plugin
            # (we can't know until we have a look at the content). So assistant will have
            # manifest-based loader. If it fails, another assistant with advanced loader will
            # be used to try again.
            dp.update({"_id": assistant.plugin_name}, {
                "$set": {
                    "plugin": {
                        "url": url,
                        "type": assistant.plugin_type,
                        "active": True
                    }
                }
            },
                upsert=True)
            assistant.handle()
            job = self.data_plugin_manager.load(assistant.plugin_name)
            assert len(job) == 1, "Expecting one job, got: %s" % job
            job = job.pop()

            def loaded(f):
                try:
                    _ = f.result()
                    self.logger.debug(
                        "Plugin '%s' downloaded, now loading manifest" %
                        assistant.plugin_name)
                    assistant.loader.load_plugin()
                except Exception as e:
                    self.logger.exception("Unable to download plugin '%s': %s" %
                                          (assistant.plugin_name, e))

            job.add_done_callback(loaded)
            return job
        else:
            raise AssistantException(
                "Could not find any assistant able to handle URL '%s'" % url)

    def load_plugin(self, plugin):
        ptype = plugin["plugin"]["type"]
        url = plugin["plugin"]["url"]
        if not plugin["plugin"]["active"]:
            self.logger.info("Data plugin '%s' is deactivated, skip" % url)
            return
        self.logger.info("Loading data plugin '%s' (type: %s)" % (url, ptype))
        if ptype in self.register:
            try:
                aklass = self.register[ptype]
                assistant = self.create_instance(aklass, url)
                assistant.handle()
                assistant.loader.load_plugin()
            except Exception as e:
                self.logger.exception("Unable to load plugin '%s': %s" %
                                      (url, e))
        else:
            raise AssistantException("Unknown data plugin type '%s'" % ptype)

    def load(self, autodiscover=True):
        """
        Load plugins registered in internal Hub database and generate/register
        dumpers & uploaders accordingly.
        If autodiscover is True, also search DATA_PLUGIN_FOLDER for existing
        plugin directories not registered yet in the database, and register
        them automatically.
        """
        plugin_dirs = []
        if autodiscover:
            try:
                plugin_dirs = os.listdir(btconfig.DATA_PLUGIN_FOLDER)
            except FileNotFoundError as e:
                raise AssistantException("Invalid DATA_PLUGIN_FOLDER: %s" % e)
        dp = get_data_plugin()
        cur = dp.find()
        for plugin in cur:
            # remove plugins from folder list if already register
            if plugin_dirs and plugin["_id"] in plugin_dirs:
                plugin_dirs.remove(plugin["_id"])
            try:
                self.load_plugin(plugin)
            except Exception as e:
                self.logger.warning("Couldn't load plugin '%s': %s" %
                                    (plugin["_id"], e))
                continue

        # some still unregistered ? (note: list always empty if autodiscover=False)
        if plugin_dirs:
            for pdir in plugin_dirs:
                fulldir = os.path.join(btconfig.DATA_PLUGIN_FOLDER, pdir)
                try:
                    self.logger.info(
                        "Found unregistered manifest-based plugin '%s', auto-register it"
                        % pdir)
                    self.register_url("local://%s" % pdir.strip().strip("/"))
                except Exception as e:
                    self.logger.exception(
                        "Couldn't auto-register plugin '%s': %s" % (pdir, e))
                    continue

    def export_dumper(self, plugin_name, folder):
        res = {
            "dumper": {
                "status": None,
                "file": None,
                "class": None,
                "message": None
            }
        }
        try:
            dclass = self.dumper_manager[plugin_name]
        except KeyError:
            res["dumper"]["status"] = "warning"
            res["dumper"][
                "message"] = "No dumper found for plugin '%s'" % plugin_name
        try:
            dumper_name = plugin_name.capitalize() + "Dumper"
            self.logger.debug("Exporting dumper %s" % dumper_name)
            assert len(dclass) == 1, "More than one dumper found: %s" % dclass
            dclass = dclass[0]
            assert hasattr(dclass, "python_code"), "No generated code found"
            dinit = os.path.join(folder, "__init__.py")
            dfile = os.path.join(folder, "dump.py")
            # clear init, we'll append code
            # we use yapf (from Google) as autopep8 (for instance) doesn't give
            # good results in term in indentation (input_type list for keylookup for instance)
            beauty, _ = yapf_api.FormatCode(dclass.python_code)
            with open(dfile, "w") as fout:
                fout.write(beauty)
            with open(dinit, "a") as fout:
                fout.write("from .dump import %s\n" % dumper_name)
            res["dumper"]["status"] = "ok"
            res["dumper"]["file"] = dfile
            res["dumper"]["class"] = dumper_name
        except Exception as e:
            res["dumper"]["status"] = "error"
            res["dumper"]["message"] = "Error exporting dumper: %s" % e
            return res

        return res

    def export_uploader(self, plugin_name, folder):
        res = {
            "uploader": {
                "status": None,
                "file": None,
                "class": None,
                "message": None
            }
        }
        try:
            uclass = self.uploader_manager[plugin_name]
        except KeyError:
            res["uploader"]["status"] = "warning"
            res["uploader"][
                "message"] = "No uploader found for plugin '%s'" % plugin_name
            return res
        try:
            uploader_name = plugin_name.capitalize() + "Uploader"
            self.logger.debug("Exporting uploader %s" % uploader_name)
            assert len(
                uclass) == 1, "More than one uploader found: %s" % uclass
            uclass = uclass[0]
            assert hasattr(uclass, "python_code"), "No generated code found"
            dinit = os.path.join(folder, "__init__.py")
            ufile = os.path.join(folder, "upload.py")
            beauty, _ = yapf_api.FormatCode(uclass.python_code)
            with open(ufile, "w") as fout:
                fout.write(beauty)
            with open(dinit, "a") as fout:
                fout.write("from .upload import %s\n" % uploader_name)
            res["uploader"]["status"] = "ok"
            res["uploader"]["file"] = ufile
            res["uploader"]["class"] = uploader_name
        except Exception as e:
            res["uploader"]["status"] = "error"
            res["uploader"]["message"] = "Error exporting uploader: %s" % e
            return res

        return res

    def export_mapping(self, plugin_name, folder):
        res = {
            "mapping": {
                "status": None,
                "file": None,
                "message": None,
                "origin": None
            }
        }
        # first check if plugin defines a custom mapping in manifest
        # if that's the case, we don't need to export mapping there
        # as it'll be exported with "uploader" code
        plugindoc = get_data_plugin().find_one({"_id": plugin_name})
        assert plugindoc, "Can't find plugin named '%s'" % plugin_name
        plugin_folder = plugindoc.get("download", {}).get("data_folder")
        assert plugin_folder, "Can't find plugin folder for '%s'" % plugin_name
        try:
            manifest = json.load(open(os.path.join(plugin_folder, "manifest.json")))
            if "mapping" in manifest.get("uploader", {}):
                res["mapping"]["message"] = "Custom mapping included in uploader export"
                res["mapping"]["status"] = "warning"
                res["mapping"]["origin"] = "custom"
                return res
        except Exception as e:
            self.logger.error("Can't read manifest while exporting code: %s" % e)
        # try to export mapping from src_master (official)
        doc = get_src_master().find_one({"_id": plugin_name})
        if doc:
            mapping = doc.get("mapping")
            res["mapping"]["origin"] = "registered"
        else:
            doc = get_src_dump().find_one({"_id": plugin_name})
            mapping = doc and doc.get("inspect", {}).get("jobs", {}).get(plugin_name, {}).get("inspect", {}).\
                get("results", {}).get("mapping")
            res["mapping"]["origin"] = "inspection"
        if not mapping:
            res["mapping"]["origin"] = None
            res["mapping"]["status"] = "warning"
            res["mapping"][
                "message"] = "Can't find registered or generated (inspection) mapping"
            return res
        else:
            ufile = os.path.join(folder, "upload.py")
            strmap, _ = yapf_api.FormatCode(pprint.pformat(mapping))
            with open(ufile, "a") as fout:
                fout.write("""
    @classmethod
    def get_mapping(klass):
        return %s\n""" % textwrap.indent((strmap), prefix="    " * 2))

        res["mapping"]["file"] = ufile
        res["mapping"]["status"] = "ok"

        return res

    def export(self,
               plugin_name,
               folder=None,
               what=["dumper", "uploader", "mapping"],
               purge=False):
        """
        Export generated code for a given plugin name, in given folder
        (or use DEFAULT_EXPORT_FOLDER if None). Exported information can be:
        - dumper: dumper class generated from the manifest
        - uploader: uploader class generated from the manifest
        - mapping: mapping generated from inspection or from the manifest
        If "purge" is true, any existing folder/code will be deleted first, otherwise,
        will raise an error if some folder/files already exist.
        """
        res = {}
        # sanity checks
        if type(what) == str:
            what = [what]
        folder = folder or self.default_export_folder
        assert os.path.exists(
            folder
        ), "Folder used to export code doesn't exist: %s" % os.path.abspath(
            folder)
        assert plugin_name  # avoid deleting the whole export folder when purge=True...
        folder = os.path.join(folder, plugin_name)
        if purge:
            rmdashfr(folder)
        if not os.path.exists(folder):
            os.makedirs(folder)
        elif not purge:
            raise FileExistsError(
                "Folder '%s' already exists, use purge=True" % folder)
        dinit = os.path.join(folder, "__init__.py")
        with open(dinit, "w") as fout:
            fout.write("")
        if "dumper" in what:
            res.update(self.export_dumper(plugin_name, folder))
        if "uploader" in what:
            res.update(self.export_uploader(plugin_name, folder))
        if "mapping" in what:
            assert "uploader" in what, "'uploader' needs to be exported too to export mapping"
            res.update(self.export_mapping(plugin_name, folder))
        # there's also at least a parser module, maybe a release module, and some more
        # dependencies, indirect, not listed in the manifest. We'll just copy everything from
        # the plugin folder to the export folder
        plugin_folder = os.path.join(btconfig.DATA_PLUGIN_FOLDER, plugin_name)
        for f in os.listdir(plugin_folder):
            src = os.path.join(plugin_folder, f)
            dst = os.path.join(folder, f)
            # useless or strictly plugin-machinery-specific, skip
            if f in ["__pycache__", "manifest.json", "__init__.py"
                     ] or f.startswith("."):
                self.logger.debug("Skipping '%s', not necessary" % src)
                continue
            self.logger.debug("Copying %s to %s" % (src, dst))
            try:
                with open(src) as fin:
                    with open(dst, "w") as fout:
                        fout.write(fin.read())
            except IsADirectoryError:
                self.logger.error(
                    "%s is a directory, expecting only files to copy" % src)
                continue

        return res



"""
    Override of elasticsearch-dsl modules to support async operations.
"""

from elasticsearch import NotFoundError, RequestError, TransportError
from elasticsearch_dsl import A, MultiSearch, Q, Search
from elasticsearch_dsl.connections import get_connection
from elasticsearch_dsl.response import Response


class AsyncMultiSearch(MultiSearch):

    async def execute(self, ignore_cache=False, raise_on_error=True):
        """
        Execute the multi search request and return a list of search results.
        """
        if ignore_cache or not hasattr(self, '_response'):
            es = get_connection(self._using)

            responses = await es.msearch(
                index=self._index,
                body=self.to_dict(),
                **self._params
            )

            out = []
            for s, r in zip(self._searches, responses['responses']):
                if r.get('error', False):
                    if raise_on_error:
                        raise TransportError('N/A', r['error']['type'], r['error'])
                    r = None
                else:
                    r = Response(s, r)
                out.append(r)

            self._response = out

        return self._response

class AsyncSearch(Search):

    async def execute(self, ignore_cache=False):
        """
        Execute the search and return an instance of ``Response`` wrapping all
        the data.

        :arg ignore_cache: if set to ``True``, consecutive calls will hit
            ES, while cached result will be ignored. Defaults to `False`
        """
        if ignore_cache or not hasattr(self, '_response'):
            es = get_connection(self._using)

            self._response = self._response_class(
                self,
                await es.search(
                    index=self._index,
                    body=self.to_dict(),
                    **self._params
                )
            )
        return self._response



"""
This module contains util functions may be shared by both BioThings data-hub and web components.
In general, do not include utils depending on any third-party modules.
Note: unittests available in biothings.tests.hub
"""
import math
import statistics
import random
import collections
import time
import re
import logging
import copy
from pprint import pformat
from datetime import datetime

import bson

from biothings.utils.common import timesofar, is_scalar, is_str, splitstr, nan, inf
from biothings.utils.web.es import flatten_doc
from biothings.utils.dataload import dict_walk


class BaseMode(object):

    # dict storing the actual specific values the mode deals with
    template = {}
    # key under which values are stored for this mode
    key = None

    def report(self, struct, drep, orig_struct=None):
        """
        Given a data structure "struct" being inspected, report (fill)
        "drep" dictionary with useful values for this mode, under
        drep[self.key] key.
        Sometimes "struct" is already converted to its analytical value at
        this point (inspect may count number of dict and would force to pass
        struct as "1", instead of the whole dict, where number of keys could be
        then be reported), "orig_struct" is that case contains the original
        structure that was to be reported, whatever the pre-conversion step did.
        """
        raise NotImplementedError("Implement in sub-class")

    def merge(self, target, tomerge):
        """
        Merge two different maps together (from tomerge into target)
        """
        raise NotImplementedError("Implement in sub-class")

    def post(self, mapt, mode, clean):
        pass

class StatsMode(BaseMode):

    template = {
        "_stats": {
            "_min": math.inf,
            "_max": -math.inf,
            "_count": 0,
            "_none": 0
        }
    }
    key = "_stats"

    def sumiflist(self, val):
        # if type(val) == list:   # TODO: remove this line
        if isinstance(val, list):
            return sum(val)
        else:
            return val

    def maxminiflist(self, val, func):
        # if type(val) == list:   # TODO: remove this line
        if isinstance(val, list):
            return func(val)
        else:
            return val

    def flatten_stats(self, stats):
        # after merge_struct, stats can be merged together as list (merge_struct
        # is only about data structures). Re-adjust here considering there could lists
        # that need to be sum'ed and min/max to be dealt with
        stats["_count"] = self.sumiflist(stats["_count"])
        stats["_max"] = self.maxminiflist(stats["_max"], max)
        stats["_min"] = self.maxminiflist(stats["_min"], min)
        return stats

    def report(self, struct, drep, orig_struct=None):
        # if is_str(struct) or type(struct) in [dict, list]:    # TODO: remove this line
        if is_str(struct) or isinstance(struct, (dict, list)):
            val = len(struct)
        else:
            val = struct
        drep[self.key]["_count"] += 1
        if val is None:
            drep[self.key]["_none"] += 1
        else:
            if val < drep[self.key]["_min"]:
                drep[self.key]["_min"] = val
            if val > drep[self.key]["_max"]:
                drep[self.key]["_max"] = val

    def merge(self, target_stats, tomerge_stats):
        target_stats = self.flatten_stats(target_stats)
        tomerge_stats = self.flatten_stats(tomerge_stats)
        # sum the counts and the sums
        target_stats["_count"] = target_stats["_count"] + tomerge_stats["_count"]
        # adjust min and max
        if tomerge_stats["_max"] > target_stats["_max"]:
            target_stats["_max"] = tomerge_stats["_max"]
        if tomerge_stats["_min"] < target_stats["_min"]:
            target_stats["_min"] = tomerge_stats["_min"]


class DeepStatsMode(StatsMode):

    template = {"_stats": {"_min": math.inf, "_max": -math.inf, "_count": 0, "__vals": []}}
    key = "_stats"

    def merge(self, target_stats, tomerge_stats):
        super(DeepStatsMode, self).merge(target_stats, tomerge_stats)
        # extend values
        target_stats.get("__vals", []).extend(tomerge_stats.get("__vals", []))

    def report(self, val, drep, orig_struct=None):
        super(DeepStatsMode, self).report(val, drep, orig_struct)
        # keep track of vals for now, stats are computed at the end
        drep[self.key]["__vals"].append(val)

    def post(self, mapt, mode, clean):
        if isinstance(mapt, dict):
            for k in list(mapt.keys()):
                if is_str(k) and k.startswith("__"):
                    if k == "__vals" and mode == "deepstats":
                        if len(mapt["__vals"]) > 1:
                            mapt["_stdev"] = statistics.stdev(mapt["__vals"])
                            mapt["_median"] = statistics.median(mapt["__vals"])
                            mapt["_mean"] = statistics.mean(mapt["__vals"])
                    if clean:
                        mapt.pop(k)
                else:
                    self.post(mapt[k], mode, clean)
        elif isinstance(mapt, list):
            for e in mapt:
                self.post(e, mode, clean)


class RegexMode(BaseMode):

    # list of {"re":"...","info":...}, if regex matches, then content
    # in "info" is used in report
    matchers = []

    def __init__(self):
        # pre-compile patterns
        for d in self.matchers:
            d["_pat"] = re.compile(d["re"])
        assert self.__class__.key, "Define class attribute 'key' in sub-class"
        self.__class__.template = {self.__class__.key: []}

    def merge(self, target, tomerge):
        # structure are lists (see template), just extend avoiding duplicated
        for e in tomerge:
            if e not in target:
                target.append(e)

    def report(self, val, drep, orig_struct=None):
        if orig_struct is not None:
            v = orig_struct
        else:
            v = val
        if is_scalar(v):
            sval = str(v)
            for dreg in self.matchers:
                if dreg["_pat"].match(sval):
                    for oneinfo in dreg["info"]:
                        if oneinfo not in drep.get(self.key, []):
                            drep.setdefault(self.key, []).append(oneinfo)


class IdentifiersMode(RegexMode):

    key = "_ident"
    # set this to a list of dict coming from http://identifiers.org/rest/collections
    ids = None

    matchers = None

    def __init__(self):
        if self.__class__.matchers is None:
            self.__class__.matchers = []
            res = {}
            # not initialized
            for ident in self.__class__.ids:
                res.setdefault(ident["pattern"], []).append(ident)
            for pat, info in res.items():
                self.__class__.matchers.append({"re": pat, "info": info})
        super().__init__()


############################################################################

MODES_MAP = {
    "stats": StatsMode,
    "deepstats": DeepStatsMode,
    "identifiers": IdentifiersMode
}

def get_mode_layer(mode):
    try:
        k = MODES_MAP[mode]
        return k()  # instance is what's used
    except KeyError:
        return None


def merge_record(target, tomerge, mode):
    mode_inst = get_mode_layer(mode)
    for k in tomerge:
        if k in target:
            if mode_inst and mode_inst.key == k:
                tgt = target[mode_inst.key]
                tom = tomerge[mode_inst.key]
                mode_inst.merge(tgt, tom)
                continue
            if not isinstance(tomerge[k], collections.Iterable):
                continue
            for typ in tomerge[k]:
                # if not an actual type we need to merge further to reach them
                if mode_inst is None and (type(typ) != type or typ == list):
                    target[k].setdefault(typ, {})
                    target[k][typ] = merge_record(target[k][typ], tomerge[k][typ], mode)
                else:
                    if mode == "type":
                        # we can safely update and possibly overwrite
                        # target with tomerge's values and in mode "type"
                        # there's no actual information for scalar fields
                        # (eg a string field will be like: {"myfield" : {str:{}}}
                        target[k].update(tomerge[k])
                    elif mode == "mapping":
                        # keep track on splitable (precedence: splitable > non-splitable)
                        # so don't merge if target has a "split" and tomerge has not,
                        # as we would loose that information
                        if splitstr is typ:
                            target.pop(k)
                            target[k] = tomerge[k]
                    elif mode_inst:
                        if typ in target[k]:
                            # same key, same type, need to merge
                            if mode_inst.key not in tomerge[k][typ]:
                                # we try to merge record at a too higher level, need to merge deeper
                                target[k] = merge_record(target[k], tomerge[k], mode)
                                continue
                            tgt = target[k][typ][mode_inst.key]
                            tom = tomerge[k][typ][mode_inst.key]
                            mode_inst.merge(tgt, tom)
                        else:
                            target[k].setdefault(typ, {}).update(tomerge[k][typ])
                    else:
                        raise ValueError("Unknown mode '%s'" % mode)
        else:
            # key doesn't exist, create key
            if mode == "type":
                target.setdefault(k, {}).update(tomerge[k])
            else:
                if mode_inst:
                    # running special mode, we just set the keys in target
                    target.setdefault(k, tomerge[k])
                else:
                    target.setdefault(k, {}).update(tomerge[k])
                # if we already have splitstr and we want to merge str, skip it
                # as splitstr > str
                if str in target and splitstr in target:
                    target.pop(str)

    return target


def inspect(struct, key=None, mapt=None, mode="type", level=0, logger=logging):
    """
    Explore struct and report types contained in it.

    Args:
        struct: is the data structure to explore
        mapt: if not None, will complete that type map with passed struct. This is useful
              when iterating over a dataset of similar data, trying to find a good type summary
              contained in that dataset.
        level: is for internal purposes, mostly debugging
        mode: see inspect_docs() documentation
    """

    mode_inst = get_mode_layer(mode)

    # init recording structure if none were passed
    if mapt is None:
        mapt = {}

    # if type(struct) == dict:    # TODO: remove this line
    if isinstance(struct, dict):
        # was this struct already explored before ? was it a list for that previous doc ?
        # then we have to pretend here it's also a list even if not, because we want to
        # report the list structure
        for k in struct:
            if mapt and list in mapt:  # and key == k:
                already_explored_as_list = True
            else:
                already_explored_as_list = False
            if False:  # already_explored_as_list:      # TODO: check this
                mapt[list].setdefault(k, {})
                typ = inspect(struct[k], key=k, mapt=mapt[list][k], mode=mode, level=level+1)
                mapt[list].update({k: typ})
            else:
                mapt.setdefault(k, {})
                typ = inspect(struct[k], key=k, mapt=mapt[k], mode=mode, level=level+1)

        if mode_inst:
            mapt.setdefault(mode_inst.key, copy.deepcopy(mode_inst.template[mode_inst.key]))
            mode_inst.report(1, mapt, struct)
    elif type(struct) == list:

        mapl = {}
        for e in struct:
            typ = inspect(e, key=key, mapt=mapl, mode=mode, level=level+1)
            mapl.update(typ)
        if mode_inst:
            # here we just report that one document had a list
            mapl.update(copy.deepcopy(mode_inst.template))
            mode_inst.report(struct, mapl)
        # if mapt exist, it means it's been explored previously but not as a list,
        # instead of mixing dict and list types, we want to normalize so we merge the previous
        # struct into that current list
        if mapt and list in mapt:
            mapt[list] = merge_record(mapt[list], mapl, mode)
        else:
            mapt.setdefault(list, {})
            mapt[list].update(mapl)
    # elif is_scalar(struct) or type(struct) == datetime:   # TODO: remove this line
    elif is_scalar(struct) or isinstance(struct, datetime):
        typ = type(struct)
        if mode == "type":
            mapt[typ] = {}
        elif mode == "mapping":
            # some type precedence processing...
            # splittable string ?
            if is_str(struct) and len(re.split(" +", struct.strip())) > 1:
                mapt[splitstr] = {}
            elif typ == bson.int64.Int64:
                mapt[int] = {}
            # we know struct is a scalar. NaN and Inf can't be indexed on ES,
            # need to catch those
            elif isinstance(struct, float) and math.isnan(struct):
                mapt[nan] = {}
            elif isinstance(struct, float) and math.isinf(struct):
                mapt[inf] = {}
            else:
                mapt[typ] = {}
            # splitstr > str
            if str in mapt and splitstr in mapt:
                mapt.pop(str)
            # float > int
            # TODO: could this be moved to es.generate_es_mapping ?
            if int in mapt and float in mapt:
                mapt.pop(int)
        else:
            mapt.setdefault(typ, copy.deepcopy(mode_inst.template))
            mode_inst.report(struct, mapt[typ])
    else:
        raise TypeError("Can't analyze type %s (data was: %s)" % (type(struct), struct))

    return mapt

def merge_scalar_list(mapt, mode):
    # TODO: this looks "strangely" to merge_record... refactoring needed ?
    # if a list is found and other keys at same level are found in that
    # list, then we need to merge. Ex: ...,{"bla":1},["bla":2],...
    mode_inst = get_mode_layer(mode)
    if "stats" in mode:
        raise NotImplementedError("merging with stats is not supported (yet)")
    if is_scalar(mapt):
        return
    if list in mapt.keys():
        other_keys = [k for k in mapt if k != list]
        for e in other_keys:
            if e in mapt[list]:
                tomerge = mapt.pop(e)
                if mode_inst:
                    for typ in tomerge:
                        if not type(typ) == type:
                            continue
                        if typ not in mapt[list][e]:
                            mapt[list][e][typ] = tomerge[typ]
                        # Note: don't update [list]["_stats" (or other modes' key)], we keep the original stats
                        # that is, what's actually been inspected on the list, originally
                        # (and we can't really update those stats as scalar stats aren't relevant
                        # to a list context
                        elif typ == mode_inst.key:
                            mode_inst.merge(mapt[list][e][mode_inst.key], tomerge[mode_inst.key])
                        else:
                            mode_inst.merge(mapt[list][e][typ][mode_inst.key], tomerge[typ][mode_inst.key])
                elif mode == "mapping":
                    for typ in tomerge:
                        if typ is str and splitstr in mapt[list][e]:
                            # precedence splitstr > str, we keep splitstr and ignore str
                            continue
                        if typ not in mapt[list][e]:
                            # that field exist in the [list] but with a different type
                            # just merge the typ
                            mapt[list][e].update(tomerge)
                        # precedence splitstr > str
                        if splitstr is typ:
                            mapt[list][e].pop(str, None)
                            mapt[list][e].update(tomerge)
                else:
                    # assuming what's in [list] is enough, we just popped the value
                    # from mapt, that's enough
                    pass
        # explore further
        merge_scalar_list(mapt[list], mode)
    # elif type(mapt) == dict:   # TODO: remove this line
    elif isinstance(mapt, dict):
        for k in mapt:
            merge_scalar_list(mapt[k], mode)
    # elif type(mapt) == list:   # TODO: remove this line
    elif isinstance(mapt, list):
        for e in mapt:
            merge_scalar_list(e, mode)


def get_converters(modes, logger=logging):
    converters = []
    # should we actually run another mode and then convert the results ?
    if "jsonschema" in modes:
        from biothings.utils.jsonschema import generate_json_schema
        # first get schema with mode="type", then convert the results
        # note "type" can't also be specified as jsonschema will replace
        # the results in _map["type"] key
        converters.append({
            "output_mode": "jsonschema",
            "input_mode": "type",
            "func": generate_json_schema,
            "delete_input_mode": "type" not in modes
        })
        modes.remove("jsonschema")
        if "type" not in modes:
            modes.append("type")

    return converters, modes


def run_converters(_map, converters, logger=logging):
    # need to convert some results ?
    for converter in converters:
        logger.info("Finalizing result for mode '%s' using converter %s", converter["output_mode"], converter)
        converted = converter["func"](_map[converter["input_mode"]])
        _map[converter["output_mode"]] = converted
        if converter["delete_input_mode"]:
            _map.pop(converter["input_mode"])

def inspect_docs(docs, mode="type", clean=True, merge=False, logger=logging,
                 pre_mapping=False, limit=None, sample=None, metadata=True,
                 auto_convert=True):
    """Inspect docs and return a summary of its structure:

    Args:
        mode: possible values are:

            - "type": (default) explore documents and report strict data structure
            - "mapping": same as type but also perform test on data so guess best mapping
                       (eg. check if a string is splitable, etc...). Implies merge=True
            - "stats": explore documents and compute basic stats (count,min,max,sum)
            - "deepstats": same as stats but record values and also compute mean,stdev,median
                         (memory intensive...)
            - "jsonschema", same as "type" but returned a json-schema formatted result

            `mode` can also be a list of modes, eg. ["type","mapping"]. There's little
            overhead computing multiple types as most time is spent on actually getting the data.
        clean: don't delete recorded vqlues or temporary results
        merge: merge scalar into list when both exist (eg. {"val":..} and [{"val":...}]
        limit: can limit the inspection to the x first docs (None = no limit, inspects all)
        sample: in combination with limit, randomly extract a sample of 'limit' docs
                (so not necessarily the x first ones defined by limit). If random.random()
                is greater than sample, doc is inspected, otherwise it's skipped
        metadata: compute metadata on the result
        auto_convert: run converters automatically (converters are used to convert one mode's
                      output to another mode's output, eg. type to jsonschema)
    """

    # if type(mode) == str:    # TODO: remove this line
    if isinstance(mode, str):
        modes = [mode]
    else:
        modes = mode
    if auto_convert:
        converters, modes = get_converters(modes, logger=logger)
    _map = {}
    for m in modes:
        _map[m] = {}
    cnt = 0
    errors = set()
    t0 = time.time()
    innert0 = time.time()

    if sample is not None:
        assert limit, "Parameter 'sample' requires 'limit' to be defined"
        assert sample != 1, "Sample value 1 not allowed (no documents would be inspected)"
    if limit:
        limit = int(limit)
        logger.debug("Limiting inspection to the %s first documents", limit)
    for doc in docs:
        if sample is not None:
            if random.random() <= sample:
                continue
        for m in modes:
            try:
                inspect(doc, mapt=_map[m], mode=m)
            except Exception as e:
                logging.exception("Can't inspect document (_id: %s) because: %s\ndoc: %s", doc.get("_id"), e, pformat("dpc"))
                errors.add(str(e))
        cnt += 1
        if cnt % 10000 == 0:
            logger.info("%d documents processed [%s]", cnt, timesofar(innert0))
            innert0 = time.time()
        if limit and cnt > limit:
            logger.debug("done")
            break
    logger.info("Done [%s]", timesofar(t0))
    logger.info("Post-processing")

    # post-process, specific for each mode
    for m in modes:
        mode_inst = get_mode_layer(m)
        if mode_inst:
            mode_inst.post(_map[m], m, clean)

    if auto_convert:
        run_converters(_map, converters, logger=logger)

    merge = "mapping" in modes and True or merge
    if merge:
        merge_scalar_list(_map["mapping"], "mapping")
    if "mapping" in modes and pre_mapping is False:
        # directly generate ES mapping
        import biothings.utils.es as es
        try:
            _map["mapping"] = es.generate_es_mapping(_map["mapping"])
            if metadata:
                # compute some extra metadata
                _map = compute_metadata(_map, "mapping")
        except es.MappingError as e:
            prem = {"pre-mapping": _map["mapping"], "errors": e.args[1]}
            _map["mapping"] = prem
    elif errors:
        _map["errors"] = errors
    return _map

def compute_metadata(mapt, mode):
    if mode == "mapping":
        flat = flatten_doc(mapt["mapping"])
        # total fields: ES6 requires to overcome the default 1000 limit if needed
        mapt["__metadata__"] = {"total_fields": len(flat)}

    return mapt


def typify_inspect_doc(dmap):
    """
    dmap is an inspect which was converted to be stored in a database,
    namely actual python types were stringify to be storabled. This function
    does the oposite and restore back python types within the inspect doc
    """
    def typify(val):
        if type(val) != type and val.startswith("__type__:"):
            typ = val.replace("__type__:", "")
            # special cases
            if typ == "NoneType":
                return None
            elif typ == "Int64":  # bson's Int64
                return bson.int64.Int64
            else:
                return eval(val.replace("__type__:", ""))
        else:
            return val
    return dict_walk(dmap, typify)


def stringify_inspect_doc(dmap):
    def stringify(val):
        if type(val) == type:
            return "__type__:%s" % val.__name__  # prevent having dots in the field (not storable in mongo)
        else:
            return str(val)
    return dict_walk(dmap, stringify)



"""
hub_db module is a place-holder for internal hub database
functions. Hub DB contains informations about sources,
configurations variables, etc... It's for internal usage.
When biothings.config_for_app() is called, this module will be
"filled" with the actual implementations from the specified backend
(speficied in config.py, or defaulting to MongoDB).

Hub DB can be implemented over different backend, it's orginally
been done using MongoDB, so the dialect is very inspired by pymongo.
Any hub db backend implementation must implement the functions and
classes below. See biothings.utils.mongo and biothings.utils.sqlit3 for
some examples.
"""
import os
import asyncio
from functools import wraps, partial

from biothings.utils.common import (dump as dumpobj, loadobj,
                                    get_random_string, get_timestamp)

def get_hub_db_conn():
    """Return a Database instance (connection to hub db)"""
    raise NotImplementedError()

def get_src_dump():
    """Return a Collection instance for src_dump collection/table"""
    raise NotImplementedError()

def get_src_master():
    """Return a Collection instance for src_master collection/table"""
    raise NotImplementedError()


def get_src_build():
    """Return a Collection instance for src_build collection/table"""
    raise NotImplementedError()


def get_src_build_config():
    """Return a Collection instance for src_build_hnonfig collection/table"""
    raise NotImplementedError()


def get_data_plugin():
    """Return a Collection instance for data_plugin collection/table"""
    raise NotImplementedError()

def get_api():
    """Return a Collection instance for api collection/table"""
    raise NotImplementedError()

def get_cmd():
    """Return a Collection instance for commands collection/table"""
    raise NotImplementedError()

def get_event():
    """Return a Collection instance for events collection/table"""
    raise NotImplementedError()

def get_hub_config():
    """Return a Collection instance storing configuration values"""
    raise NotImplementedError()

def get_last_command():
    """Return the latest cmd document (according to _id)"""
    raise NotImplementedError()

def get_source_fullname(col_name):
    """
    Assuming col_name is a collection created from an upload process,
    find the main source & sub_source associated.
    """
    raise NotImplementedError()

class IDatabase(object):
    """
    This class declares an interface and partially implements some of it,
    mimicking mongokit.Connection class. It's used to keep used document model.
    Any internal backend should implement (derives) this interface
    """

    def __init__(self):
        super(IDatabase, self).__init__()
        self.name = None   # should be set from config module
        # any other initialization can be done here,
        # depending on the backend specifics

    @property
    def address(self):
        """Returns sufficient information so a connection to a database
        can be created. Information can be a dictionary, object, etc...
        and depends on the actual backend"""
        raise NotImplementedError()

    #TODO: really needed ? on is it for src_db only ?
    def collection_names(self):
        """Return a list of all collections (or tables) found in this database"""
        raise NotImplementedError()

    def create_collection(self, colname):
        """Create a table/colleciton named colname. If backend is using a schema-based
        database (ie. SQL), backend should enforce the schema with at least field "_id"
        as the primary key (as a string)."""
        raise NotImplementedError()

    def __getitem__(self, colname):
        """Return a Collection object for colname
        (and create it if it doesn't exist)"""
        raise NotImplementedError()

    def __repr__(self):
        return "<%s at %s, %s>" % (self.__class__.__name__, hex(id(self)), self.address)


class Collection(object):
    """
    Defines a minimal subset of MongoDB collection behavior.
    Note: Collection instances must be pickleable (if not, __getstate__
    can be implemented to deal with those attributes for instance)
    """

    def __init__(self, colname, db):
        """Init args can differ depending on the backend requirements.
           colname is the only one required."""
        self.colname = colname
        raise NotImplementedError()

    @property
    def name(self):
        """Return the collection/table name"""
        return self.colname

    def database(self):
        """Return the database name"""
        raise NotImplementedError()

    def find_one(self, *args, **kwargs):
        """Return one document from the collection. `*args` will contain
        a dict with the query parameters. See also find()"""
        raise NotImplementedError()

    def find(self, *args, **kwargs):
        """Return an iterable of documents matching criterias defined in
        `*args[0]` (which will be a dict). Query dialect is a minimal one, inspired
        by MongoDB. Dict can contain the name of a key, and the value being searched for.
        Ex: {"field1":"value1"} will return all documents where field1 == "value1".
        Nested key (field1.subfield1) aren't supported (no need to implement).
        Exact matches only are required.

        If no query is passed, or if query is an empty dict, return all documents.
        """
        raise NotImplementedError()

    def insert_one(self, doc):
        """Insert a document in the collection. Raise an error if already inserted"""
        raise NotImplementedError()

    def update_one(self, query, what, upsert=False):
        """Update one document (or the first matching query). See find() for query parameter.
        "what" tells how to update the document. $set/$unset/$push operators must be implemented
        (refer to MongoDB documentation for more). Nested keys operation aren't necesary.
        """
        raise NotImplementedError()

    def update(self, query, what):
        """Same as update_one() but operate on all documents matching 'query'"""
        raise NotImplementedError()

    def save(self, doc):
        """Shortcut to update_one() or insert_one(). Save the document, by
        either inserting if it doesn't exist, or update existing one"""
        raise NotImplementedError()

    def replace_one(self, query, doc):
        """Replace a document matching 'query' (or the first found one) with passed doc"""
        raise NotImplementedError()

    def remove(self, query):
        """Delete all documents matching 'query'"""
        raise NotImplementedError()

    def count(self):
        """Return the number of documents in the collection"""
        raise NotImplementedError()

    def __getitem__(self, _id):
        """Shortcut to find_one({"_id":_id})"""
        raise NotImplementedError()

def backup(folder=".", archive=None):
    """
    Dump the whole hub_db database in given folder. "archive" can be pass
    to specify the target filename, otherwise, it's randomly generated

    .. note:: this doesn't backup source/merge data, just the internal data
             used by the hub
    """
    # get database name (ie. hub_db internal database)
    db_name = get_src_dump().database.name
    dump = {}
    for getter in [get_src_dump, get_src_master, get_src_build,
                   get_src_build_config, get_data_plugin, get_api,
                   get_cmd, get_event, get_hub_config]:
        col = getter()
        dump[col.name] = []
        for doc in col.find():
            dump[col.name].append(doc)
    if not archive:
        archive = "backup_%s_%s.pyobj" % (get_timestamp(), get_random_string())
    path = os.path.join(folder, archive)
    dumpobj(dump, path)
    return path

def restore(archive, drop=False):
    """Restore database from given archive. If drop is True, then delete existing collections"""
    data = loadobj(archive)
    # use src_dump collection which always exists to get the database object
    db = get_src_dump().database
    for colname in data:
        docs = data[colname]
        col = db[colname]
        if drop:
            # we don't have a drop command but we can remove all docs
            col.remove({})
        for doc in docs:
            col.save(doc)

#########################################################################
# small pubsub framework to track changes in hub db internal collection #
#########################################################################


class ChangeListener(object):

    def read(self):
        raise NotImplementedError("Implement me")


class ChangeWatcher(object):

    listeners = set()
    event_queue = asyncio.Queue()
    do_publish = False

    col_entity = {
        "src_dump": "source",
        "src_build": "build",
        "src_build_config": "build_config",
        "src_master": "master",
        "cmd": "command",
        "hub_config": "config",
    }

    @classmethod
    def publish(cls):
        cls.do_publish = True
        @asyncio.coroutine
        def do():
            while cls.do_publish:
                evt = yield from cls.event_queue.get()
                for listener in cls.listeners:
                    try:
                        listener.read(evt)
                    except Exception as e:
                        pass
                        #logging.error("Can't publish %s to %s: %s" % (evt,listener,e))
        return asyncio.ensure_future(do())

    @classmethod
    def add(cls, listener):
        assert hasattr(listener, "read"), "Listener '%s' has no read() method" % listener
        cls.listeners.add(listener)
        cls.publish()

    @classmethod
    def monitor(cls, func, entity, op):
        @wraps(func)
        def func_wrapper(*args, **kwargs):
            # don't speak alone in the immensity of the void
            if cls.listeners:
                # try to narrow down the event to a doc
                # analyse the query/filter (1st elem in args), it tells how many docs are
                # impacted, thus telling us wether to send a detailed or general event
                # if args and type(args[0]) == dict and "_id" in args[0]:     # TODO: replaced this line, delete it after confirmed.
                if args and isinstance(args[0], dict) and "_id" in args[0]:
                    # single event associated to one ID, we send an "detailed" event
                    event = {"_id": args[0]["_id"], "obj": entity, "op": op}
                    if entity == "event":
                        # sends everything
                        event["data"] = args[0]
                    cls.event_queue.put_nowait(event)
                else:
                    # can't find ID, we send a general event (not specific to one doc)
                    event = {"obj": entity, "op": op}
                    cls.event_queue.put_nowait(event)

            return func(*args, **kwargs)
        return func_wrapper

    @classmethod
    def wrap(cls, getfunc):
        def decorate():
            col = getfunc()
            for method in ["insert_one", "update_one", "update",
                           "save", "replace_one", "remove"]:
                colmethod = getattr(col, method)
                colname = getfunc.__name__.replace("get_", "")
                colmethod = cls.monitor(
                    colmethod,
                    entity=cls.col_entity.get(colname, colname),
                    op=method
                )
                setattr(col, method, colmethod)
            return col
        return partial(decorate)


def setup(config):
    global get_hub_db_conn
    global get_src_dump
    global get_src_master
    global get_src_build
    global get_src_build_config
    global get_data_plugin
    global get_api
    global get_cmd
    global get_event
    global get_hub_config
    global get_source_fullname
    global get_last_command
    get_hub_db_conn = config.hub_db.get_hub_db_conn
    # use ChangeWatcher on internal collections so we can publish changes in real-time
    get_src_dump = ChangeWatcher.wrap(config.hub_db.get_src_dump)
    get_src_master = ChangeWatcher.wrap(config.hub_db.get_src_master)
    get_src_build = ChangeWatcher.wrap(config.hub_db.get_src_build)
    get_src_build_config = ChangeWatcher.wrap(config.hub_db.get_src_build_config)
    get_data_plugin = ChangeWatcher.wrap(config.hub_db.get_data_plugin)
    get_api = ChangeWatcher.wrap(config.hub_db.get_api)
    get_cmd = ChangeWatcher.wrap(config.hub_db.get_cmd)
    get_event = ChangeWatcher.wrap(config.hub_db.get_event)
    get_hub_config = ChangeWatcher.wrap(config.hub_db.get_hub_config)
    get_source_fullname = config.hub_db.get_source_fullname
    get_last_command = config.hub_db.get_last_command
    # propagate config module to classes
    config.hub_db.Database.CONFIG = config



import sys
import os
import asyncio
import types
import copy
import inspect
import importlib
import re
import json
import logging
import concurrent.futures
from .version import MAJOR_VER, MINOR_VER, MICRO_VER
from .utils.dotfield import merge_object, make_object
from .utils.jsondiff import make as jsondiff
from .utils.dataload import dict_traverse


def get_version():
    return '{}.{}.{}'.format(MAJOR_VER, MINOR_VER, MICRO_VER)


# re pattern to find config param
# (by convention, all upper caps, _ allowed, that's all)
PARAM_PAT = re.compile("^([A-Z_]+)$")


class ConfigurationError(Exception):
    pass


class ConfigurationValue(object):
    """
    type to wrap default value when it's code and needs to be interpreted later
    code is passed to eval() in the context of the whole "config" dict
    (so for instance, paths declared before in the configuration file can be used
    in the code passed to eval)
    code will also be executed through exec() *if* eval() raised a syntax error. This
    would happen when code contains statements, not just expression. In that case,
    a variable should be created in these statements (named the same as the original
    config variable) so the proper value can be through ConfigurationManager.
    """
    def __init__(self, code):
        self.code = code

    def get_value(self, name, conf):
        """
        Return value by eval'ing code in self.code, in the context of given configuration
        dict (namespace), for given config parameter name.
        """
        try:
            return eval(self.code, conf.__dict__)
        except SyntaxError:
            # try exec, maybe it's a statement (not just an expression).
            # in that case, it eeans user really knows what he's doing...
            exec(self.code, conf.__dict__)
            # there must be a variable named the same same, in that dict,
            # coming from code's statements
            return conf.__dict__[name]


class ConfigurationDefault(object):
    def __init__(self, default, desc):
        self.default = default
        self.desc = desc


def check_config(config_mod):
    for attr in dir(config_mod):
        if isinstance(getattr(config_mod, attr), ConfigurationError):
            raise ConfigurationError("%s: %s" %
                                     (attr, str(getattr(config_mod, attr))))


class ConfigurationManager(types.ModuleType):
    """
    Wraps and manages configuration access and edit. A singleton
    instance is available throughout all hub apps using biothings.config
    after biothings.config_for_app(conf_mod) as been called.
    In addition to providing config value access, either from config files
    or database, config manager can supersede attributes of a class with values
    coming from the database, allowing dynamic configuration of hub's elements.
    """
    def __init__(self, conf):
        self.conf = conf
        self.conf_parser = ConfigParser(conf)
        self.hub_config = None  # collection containing config value, set when wrapped, see config_for_app()
        self.bykeys = {}  # caching values from hub db
        self.byroots = {
        }  # for dotfield notation, all config names starting with first elem
        self.allow_edits = False
        self.dirty = False  # gets dirty (needs reload) when config is changed in db
        self._original_params = {
        }  # cache config params as defined in config files
        self.setup()

    @property
    def original_params(self):
        if not self._original_params:
            self._original_params = self.get_config_params()
        return self._original_params

    def __getattr__(self, name):
        # first try value from Hub DB, they have precedence
        # if nothing, then take it from file
        try:
            val = self.get_value_from_db(name)
        except (KeyError, ValueError):
            val = self.get_value_from_file(name)

        return val

    def __delattr__(self, name):
        delattr(self.conf, name)

    def __getitem__(self, name):
        # for dotfield notation
        return self.__getattr__(name)

    def setup(self):
        can_edit = False
        if hasattr(self.conf, "CONFIG_READONLY"):
            can_edit = not self.conf.CONFIG_READONLY
            delattr(self.conf, "CONFIG_READONLY")
        self.allow_edits = bool(can_edit)

    def show(self):
        origparams = copy.deepcopy(self.original_params)
        byscopes = {"scope": {}}
        # some of these could have been superseded by values from DB
        for key, info in origparams.items():
            # search whether param named "key" has been superseded
            if info["default"] != info["value"]:
                diff = jsondiff(info["default"], info["value"])
                origparams[key]["diff"] = diff
        byscopes["scope"]["config"] = origparams
        byscopes["scope"]["class"] = self.bykeys.get("class", {})

        # set a flag to indicate if config is dirty and the hub needs to reload
        # (something's changed but not taken yet into account)
        byscopes["_dirty"] = self.dirty
        byscopes["allow_edits"] = self.allow_edits
        return byscopes

    def clear_cache(self):
        self.bykeys = {}
        self.byroots = {}
        self._original_params = {}

    def get_path_from_db(self, name):
        return self.byroots.get(name, [])

    def merge_with_path_from_db(self, name, val):
        roots = self.get_path_from_db(name)
        for root in roots:
            dotfieldname, value = root["_id"], root["value"]
            val = merge_object(
                val,
                make_object(dotfieldname, value)[dotfieldname.split(".")[0]])
        return val

    def check_editable(self, name, scope):
        assert self.allow_edits, "Configuration is read-only"
        assert self.hub_config, "No hub_config collection set"
        assert not name == "CONFIG_READONLY", "I won't allow to store/supersede that parameter. Nice try though..."
        # check if param is invisble
        # (and even if using dotfield notation, if a dict is "invisible"
        # then any of its keys are also invisible)
        if scope == "config":
            name = name.split(".")[0]
            assert name in self.original_params, "Unknown configuration parameter"
            assert not self.original_params[name].get(
                "readonly"), "This parameter is not editable"

    def reset(self, name, scope="config"):
        self.check_editable(name, scope)
        res = self.hub_config.remove({"_id": name})
        self.dirty = True  # may need a reload
        self.clear_cache(
        )  # will force reload everything to get up-to-date values
        return res["ok"]

    def store_value_to_db(self, name, value, scope="config"):
        """
        Stores a configuration "value" named "name" in hub_config.
        "scope" defines what the configuration value applies on:
        - 'config': a config value which could be find in config*.py files
        - 'class': applied to a class (supersedes class attributes)
        """
        self.check_editable(name, scope)
        res = self.hub_config.update_one(
            {"_id": name},
            {"$set": {
                "scope": scope,
                "value": json.loads(value),
            }},
            upsert=True)
        self.dirty = True  # may need a reload
        self.clear_cache()
        return res

    def get_value_from_db(self, name, scope="config"):
        if self.hub_config:
            # cache on first call
            if not self.bykeys:
                for d in self.hub_config.find():
                    # tricky: get it from file to cast to correct type
                    val = d["value"]
                    try:
                        tval = self.get_value_from_file(d["_id"])
                        typ = type(tval)
                        val = typ(val)  # recast
                    except AttributeError:
                        # only exists in db
                        pass
                    # fill in cache, by scope then by config key
                    scope = d.get("scope", "config")
                    self.bykeys.setdefault(scope, {})
                    self.bykeys[scope][d["_id"]] = val
                    elems = d["_id"].split(".")
                    if len(elems) > 1:  # we have a dotfield notation there
                        # tricky; read the comments below, extracting the root has a different meaning depending on the scope
                        if scope == "config":
                            # first elem in the path a config variable, the rest is a path inside that variable (which
                            # is a dict)
                            self.byroots.setdefault(elems[0], []).append({
                                "_id":
                                d["_id"],
                                "value":
                                val
                            })
                        else:
                            # the root is everything up to the last element in the path, that is, the full path
                            # of the class, etc... The last element is the attribute to set.
                            self.byroots.setdefault(".".join(elems[:-1]),
                                                    []).append({
                                                        "_id": d["_id"],
                                                        "value": val
                                                    })

            return self.bykeys.get(scope, {})[name]
        else:
            raise ValueError("hub_config not set yet")

    def get_value_from_file(self, name):
        # if "name" corresponds to a dict, we may have
        # dotfield paths in DB overridiing some of the content
        # we'd need to merge that path with
        val = getattr(self.conf, name)
        try:
            copiedval = copy.deepcopy(
                val)  # we want to keep original value (if it's a dict)
            # as this will be merged with value from db
        except TypeError:
            # it can't be copied, it probably means it can't even be stored in db
            # so no risk of overriding original value
            copiedval = val
            pass
        copiedval = self.merge_with_path_from_db(name, copiedval)

        def eval_default_value(k, v):
            if isinstance(v, ConfigurationDefault):
                if isinstance(v.default, ConfigurationValue):
                    return (k, v.default.get_value(name, self.conf))
                else:
                    return (k, v.default)
            elif isinstance(v, ConfigurationValue):
                return (k, v.get_value(k, self.conf))
            else:
                return (k, v)

        if isinstance(copiedval, dict):
            # walk the dict and instantiate values when special
            dict_traverse(copiedval, eval_default_value, traverse_list=True)
        else:
            # just use the same func but ignore "k" key, not a dict
            # pass unhashable "k" to make sure we'd raise an error
            # while dict traversing  if we're not supposed to be here
            _, copiedval = eval_default_value({}, copiedval)

        return copiedval

    def patch(self, something, confvals, scope):
        for confval in confvals:
            key = confval["_id"]
            value = confval["value"]
            # key looks like dotfield notation, last elem is the attribute, and what's before
            # is path to the "something" object (eg. hub.dataload.sources.mysrc.dump.Dumper)
            attr = key.split(".")[-1]
            setattr(something, attr, value)

    def get_config_params(self):
        """
        Return all configuration parameters with their description, by section"
        """
        params = {}
        for attrname in self.conf_parser.list():
            value = getattr(self, attrname)
            info = self.conf_parser.find_information(attrname)
            if info is None:
                # if no information could be found, not even the field,
                # (if field was found in a config file but without any information,
                # we would have had a {"found" : True}), it means the parameter (field)
                # as been set dynamically somewhere in the code. It's not coming from
                # config files, we need to tag it as-is, and make it readonly
                # (we don't want to allow config changes other than those specified in
                # config files)
                params[attrname] = {
                    "value": value,
                    "dynamic": True,
                    "readonly": True,
                    "default": value  # for compatibilty
                }
            else:
                if info["invisible"]:
                    continue
                origvalue = getattr(info["confmod"], attrname)
                params[attrname] = {
                    "value": info["hidden"] and "********" or value,
                    "section": info["section"],
                    "desc": info["desc"],
                    "default": info["hidden"] and "********" or origvalue
                }
                if info["hidden"]:
                    params[attrname]["hidden"] = True
                if info["readonly"]:
                    params[attrname]["readonly"] = True

        return params

    def supersede(self, something):
        # find config values with scope corresponding to something's type
        # Note: any conf key will look like a dotfield notation, eg. MyClass.myattr
        # so we search 1st by root (MyClass) to see if we have a config key in DB,
        # then we fetch all
        scope = None
        if isinstance(something, type):
            fullname = "%s.%s" % (something.__module__, something.__name__)
            scope = "class"
        else:
            raise TypeError("Don't know how to supersede type '%s'" %
                            type(something))

        assert scope
        # it's a class, get by roots using string repr
        confvals = self.byroots.get(fullname, [])
        # check/filter by scope
        valids = []
        for conf in confvals:
            match = self.get_value_from_db(conf["_id"], scope=scope)
            if match:
                # we actually have a conf key/value matching that scope, keep it
                valids.append(conf)
            self.patch(something, valids, scope)

    def __repr__(self):
        return "<%s over %s>" % (self.__class__.__name__, self.conf.__name__)


class ConfigParser(object):
    """
    Parse configuration module and extract documentation from it.
    Documentation can be found in different place (in order):
    1. the configuration value is a ConfigurationDefault instance (specify a default value)
       or a ConfigurationError instance, in whic case the documentation is taken
       from the instance doc.
    2. the documentation can be specified as an inline comment
    3. the documentation can be specified as comments above

    If the configuration module also import another (or more) config modules, those
    modules will be searched as well, if nothing could be found in the main module.
    As soon as a documentation is found, the search stops (importance of module imports order)

    There are several special comment formats used to organize and manager params:
    - all comments above a param are considered as documentation/description for the parameter,
      until a empty line or a non-comment line is found.
    - to be considered as documentation, comments also need to have a space after "#". So:
            # my comment
      will be kepts as documentation, but:
            #my comment
      will just be ignored (so python code can be commented that way, without then being part
      of the documentation)
    - A section title can be added to organize parameters. A section title looks like this:
            #* my section title *#
      It can be added for each parameters, or the first section found above is the section
      the parameter will be associated to. An empty section title can be used to reset the
      section title and associate the current parameter to default one:
            #* *#
      If no section is found, all parameters are part of the default one (None).
    - some parameters needs to be kept secret (like passwords for instance):
            #- invisible -#
      will hide the parameter, including the name, value, description, from the configuration
            #- hidden -#
      will keep the parameter in the configuration displayed to users, but its value will be omitted
            #- readonly -#
      will allow the parameter to shown, but not editable

      Note: special comments can stacked:
            #- readonly -#
            #- hidden -#
      will make the parameter read-only, and its value won't be displayed
    """
    def __init__(self, config_mod):
        self.config = config_mod
        self.lines = inspect.getsourcelines(self.config)[0]
        self.find_base_config()

    def list(self):
        """
        Return a list of all config parameters' names found in config file
        (including base config files)
        """
        for attrname in dir(self.config):
            if PARAM_PAT.match(attrname):
                yield attrname

    def find_base_config(self):
        self.config_bases = []
        pat = re.compile(r"^from\s+(.*?)\s+import\s+\*")
        for l in self.lines:
            m = pat.match(l)
            if m:
                base = m.groups()[0]
                base_mod = importlib.import_module(base)
                self.config_bases.append(base_mod)

    def find_information(self, field):
        # search all config files, trying to get max infor
        # (field can be set in a base config file containing the description
        # and re-defined in main config without description)
        infos = []
        for conf in [self.config] + self.config_bases:
            info = self.find_docstring_in_config(conf, field)
            if info["found"]:
                info["confmod"] = conf
                infos.insert(0, info)

        if infos:
            # merge everything we have about the field, in the order
            # so most recent config in import history has precedence
            master = infos[0]
            for info in infos[1:]:
                for k, v in info.items():
                    if v:
                        master[k] = v
            return master

        # if we get there, it means field couldn't be found in any config files
        return None

    def find_docstring_in_config(self, config, field):
        field = field.strip()
        if "." in field:
            # dotfield notation, explore a dict
            raise NotImplementedError("docstring no supported in dict")
            pass
        if not hasattr(config, field):
            return {"found": False}
        confval = getattr(config, field)
        desc = None
        section = None
        invisible = None
        hidden = None
        readonly = None
        if isinstance(confval, ConfigurationDefault):
            desc = confval.desc
        if isinstance(confval, ConfigurationError):
            # it's an Exception, just take the text
            desc = confval.args[0]

        found_field = False
        lines = inspect.getsourcelines(config)[0]
        for i, l in enumerate(lines):
            if l.startswith(field):
                found_field = True
                break
        if found_field:
            section = self.find_section(field, lines, i)
            invisible = self.is_invisible(field, lines, i)
            hidden = self.is_hidden(field, lines, i)
            readonly = self.is_readonly(field, lines, i)
            if not desc:
                # no desc previously found in ConfigurationDefault or ConfigurationError
                desc = self.find_description(field, lines, i)
        return {
            "found": found_field,
            "section": section,
            "desc": desc,
            "invisible": invisible,
            "hidden": hidden,
            "readonly": readonly
        }

    def find_description(self, field, lines, lineno):
        # at least one space after # to consider this a description
        descpat = re.compile(r".*\s*#\s+(.*)$")
        # inline comment
        line = lines[lineno]
        m = descpat.match(line)
        if m:
            return m.groups()[0].strip()
        else:
            # comment above
            cmts = []
            i = lineno
            while i > 0:
                i -= 1
                line = lines[i]
                if line.startswith("\n"):
                    break
                else:
                    m = descpat.match(line)
                    if m:
                        cmts.insert(0, m.groups()[0])
                    else:
                        break
            # filter out empty lines
            cmts = [c for c in cmts if c]
            if cmts:
                return "\n".join(cmts)
            else:
                return None

    def find_special_comment(self,
                             pattern,
                             field,
                             lines,
                             lineno,
                             stop_on_eol=True):
        pat = re.compile(pattern)
        commentpat = re.compile(r"^\s*#")
        i = lineno
        while i > 0:
            i -= 1
            line = lines[i]
            if stop_on_eol and not commentpat.match(line):
                # not even a comment, brek
                return False
            if stop_on_eol and line.startswith("\n"):
                break
            m = pat.match(line)
            if m:
                return m

    def find_section(self, field, lines, lineno):
        match = self.find_special_comment(r"^#\*\s*(.*)\s*\*#\s*$",
                                          field,
                                          lines,
                                          lineno,
                                          stop_on_eol=False)
        if match:
            section = match.groups()[0]
            if not section:  # if we have "#* *#" it's a section breaker
                section = None  # back to None if empty string
            else:
                section = section.strip()
            return section

    def is_invisible(self, field, lines, lineno):
        match = self.find_special_comment(r"^#-\s*invisible\s*-#\s*$", field,
                                          lines, lineno)
        if match:
            return True
        else:
            return False

    def is_hidden(self, field, lines, lineno):
        match = self.find_special_comment(r"^#-\s*hide\s*-#\s*$", field, lines,
                                          lineno)
        if match:
            return True
        else:
            return False

    def is_readonly(self, field, lines, lineno):
        match = self.find_special_comment(r"^#-\s*readonly\s*-#\s*$", field,
                                          lines, lineno)
        if match:
            return True
        else:
            return False


def config_for_app(config_mod, check=True):
    if check is True:
        check_config(config_mod)
    app_path = os.path.split(config_mod.__file__)[0]
    sys.path.insert(0, app_path)
    # this will create a "biothings.config" module
    # so "from biothings from config" will get app config at lib level
    # (but "import biothings.config" won't b/c not a real module within biothings
    wrapper = ConfigurationManager(config_mod)
    globals()["config"] = wrapper
    config.APP_PATH = app_path
    if not hasattr(config_mod, "HUB_DB_BACKEND"):
        raise AttributeError(
            "Can't find HUB_DB_BACKEND in configutation module")
    else:
        import importlib
        config.hub_db = importlib.import_module(
            config_mod.HUB_DB_BACKEND["module"])
        import biothings.utils.hub_db
        biothings.utils.hub_db.setup(config)
        wrapper.hub_config = biothings.utils.hub_db.get_hub_config()
    from biothings.utils.loggers import EventRecorder
    logger = logging.getLogger()
    fmt = logging.Formatter(
        '%(asctime)s [%(process)d:%(threadName)s] - %(name)s - %(levelname)s -- %(message)s',
        datefmt="%H:%M:%S")
    erh = EventRecorder()
    erh.name = "event_recorder"
    erh.setFormatter(fmt)
    if erh.name not in [h.name for h in logger.handlers]:
        logger.addHandler(erh)


def get_loop(max_workers=None):
    loop = asyncio.get_event_loop()
    executor = concurrent.futures.ProcessPoolExecutor(max_workers=max_workers)
    loop.set_default_executor(executor)
    return loop


# adjust some loggers...
if os.environ.get("HUB_VERBOSE","0") != "1":
    logging.getLogger("elasticsearch").setLevel(logging.ERROR)
    logging.getLogger("urllib3").setLevel(logging.ERROR)
    logging.getLogger("requests").setLevel(logging.ERROR)
    logging.getLogger("botocore").setLevel(logging.ERROR)
    logging.getLogger("boto").setLevel(logging.ERROR)
    logging.getLogger("git").setLevel(logging.ERROR)



"""
Biothings Web Handlers

biothings.web.handlers.BaseHandler

    Supports:
    - access to biothings.web.settings instance
    - access to biothings.web.handlers logging stream
    - access to default web templates
    - Sentry error monitoring

    Subclasses:
    - biothings.web.handlers.StatusHandler
    - biothings.web.handlers.FrontPageHandler
    - discovery.web.handlers.BaseHandler
    ...

Also available:

biothings.web.handlers.BaseAPIHandler
biothings.web.handlers.BaseESRequestHandler
biothings.web.handlers.ESRequestHandler

"""
import copy
import logging

from elasticsearch.exceptions import ElasticsearchException
from tornado.web import RequestHandler

try:
    from raven.contrib.tornado import SentryMixin
except ImportError:
    class SentryMixin(object):
        """dummy class mixin"""

class BaseHandler(SentryMixin, RequestHandler):
    """
        Parent class of all handlers, only direct descendant of `tornado.web.RequestHandler
        <http://www.tornadoweb.org/en/stable/web.html#tornado.web.RequestHandler>`_,
    """

    @property
    def web_settings(self):
        try:
            setting = self.settings['biothings']
        except KeyError:
            self.require_setting('biothings')
        else:
            return setting

    def get_sentry_client(self):
        """
        Override default and retrieve from tornado setting instead.
        """
        client = self.settings.get('sentry_client')
        if not client:
            self.require_setting('sentry_client')
        return client

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def log_exception(self, *args, **kwargs):
        """
        Only attempt to report to Sentry when the client is setup.
        Discard when API key is not set or raven is not installed.
        """
        if 'sentry_client' in self.settings:
            return SentryMixin.log_exception(self, *args, **kwargs)
        return RequestHandler.log_exception(self, *args, **kwargs)

    def get_template_path(self):
        if "template_path" in self.settings:
            return super().get_template_path()
        import biothings.web.templates
        return next(iter(biothings.web.templates.__path__))

    def data_received(self, chunk):
        """
        Implement this method to handle streamed request data.
        """
        # this dummy implementation silences the pylint abstract-class error

class StatusHandler(BaseHandler):
    '''
    Handles requests to check the status of the server.
    Use set_status instead of raising exception so that
    no error will be propogated to sentry monitoring.
    '''

    def head(self):
        return self._check()

    async def get(self):
        res = await self._check()
        self.finish(res)

    async def _check(self):

        client = self.web_settings.async_es_client
        payload = self.web_settings.STATUS_CHECK

        status = None  # green, red, yellow
        res = None  # additional doc check

        try:
            health = await client.cluster.health()
            status = health['status']
            if payload:
                res = await client.get(**payload)

        except ElasticsearchException:
            self.set_status(503)

        else:
            if status == 'red':
                self.set_status(503)
            if payload and not res:
                self.set_status(503)

        return {
            "code": self.get_status(),
            "status": status,
            "payload": payload,
            "response": res
        }


class FrontPageHandler(BaseHandler):

    def get(self):

        self.render(
            template_name="home.html",
            alert='Front Page Not Configured.',
            title='Biothings API',
            contents=self.web_settings.handlers.keys(),
            support=self.web_settings.BIOTHING_TYPES,
            url='http://biothings.io/'
        )


from .api import *
from .es import *



import logging
from collections import OrderedDict

import elasticsearch

from biothings.utils.common import is_seq
from biothings.utils.doc_traversal import depth_first_traversal


async def get_es_versions(client):
    es_version = 'unknown'
    es_cluster = 'unknown'
    try:
        info = await client.info(request_timeout=3)
        version = info['version']['number']
        cluster = info['cluster_name']
        health = await client.cluster.health(request_timeout=3)
        status = health['status']
    except elasticsearch.TransportError as exc:
        logger = logging.getLogger(__name__)
        logger.error('Error reading elasticsearch status.')
        logger.debug(exc)
    else:
        es_version = version
        es_cluster = f"{cluster} ({status})"
    return {
        "elasticsearch_version": es_version,
        "elasticsearch_cluster": es_cluster,
    }

def exists_or_null(doc, field, val=None):
    def _helper(doc, _list, val):
        if isinstance(doc, dict):
            if len(_list) > 1:
                if _list[0] not in doc:
                    doc[_list[0]] = {}
                _helper(doc[_list[0]], _list[1:], val)
            else:
                if _list[0] not in doc:
                    doc[_list[0]] = val
        elif is_seq(doc):
            for o in doc:
                _helper(o, _list, val)

    _helper(doc, list(field.split('.')), val)

    return doc


def flatten_doc_2(doc, outfield_sep='.', sort=True):
    _ret = {}
    for _path, _val in depth_first_traversal(doc):
        if not isinstance(_val, dict) and not is_seq(_val):
            if outfield_sep:
                _new_path = outfield_sep.join(_path)
            else:
                _new_path = _path
            _ret.setdefault(_new_path, []).append(_val)
    if sort and outfield_sep:
        return OrderedDict(sorted([(k, v[0]) if len(v) == 1 else (k, v) for (k, v) in _ret.items()], key=lambda x: x[0]))
    return dict([(k, v[0]) if len(v) == 1 else (k, v) for (k, v) in _ret.items()])


def flatten_doc(doc, outfield_sep='.', sort=True):
    ''' This function will flatten an elasticsearch document (really any json object).
        outfield_sep is the separator between the fields in the return object.
        sort specifies whether the output object should be sorted alphabetically before returning
            (otherwise output will remain in traveral order) '''

    def _recursion_helper(_doc, _ret, out):
        if isinstance(_doc, dict):
            for key in _doc:
                if outfield_sep:
                    new_key = key if not out else outfield_sep.join([out, key])
                else:
                    new_key = tuple([key]) if not out else tuple(list(tuple(out)) + [key])
                _recursion_helper(_doc[key], _ret, new_key)
        elif is_seq(_doc):
            for _obj in _doc:
                _recursion_helper(_obj, _ret, out)
        else:
            # this is a leaf
            _ret.setdefault(out, []).append(_doc)

    ret = {}
    _recursion_helper(doc, ret, '')
    if sort and outfield_sep:
        return OrderedDict(sorted([(k, v[0]) if len(v) == 1 else (k, v) for (k, v) in ret.items()], key=lambda x: x[0]))
    return dict([(k, v[0]) if len(v) == 1 else (k, v) for (k, v) in ret.items()])



#!/usr/bin/env python
import os
import argparse
from string import Template

import biothings


settings_dict = {
    "src_package": "",
    "settings_class": "",
    "annotation_endpoint": "",
    "query_endpoint": "",
    "annotation_handler_name": "",
    "query_handler_name": "",
    "es_doctype": "",
    "base_url": "",
    "nosetest_settings_class": "",
    "nosetest_envar": ""
}


def main(args):
    # now only es is supported as a biothing backend...
    if args.b == 'es':
        subdir = 'es_project_template'
    else:
        pass

    # get the template dir for the proper backend
    template_dir = os.path.join(os.path.split(biothings.__file__)[0], 'conf', 'biothings_templates', subdir)

    if args.verbose:
        print("Template directory: {}".format(template_dir))

    # generate template settings
    settings_dict["src_package"] = 'my' + args.obj.lower()
    settings_dict["settings_class"] = "My" + args.obj.title() + "WebSettings"
    settings_dict["annotation_endpoint"] = args.obj.lower()
    settings_dict["query_endpoint"] = "query"
    settings_dict["annotation_handler_name"] = args.obj.title() + "Handler"
    settings_dict["query_handler_name"] = "QueryHandler"
    settings_dict["es_doctype"] = args.obj.lower()
    settings_dict["base_url"] = "My" + args.obj.title() + ".info"
    settings_dict["nosetest_settings_class"] = args.obj.title()
    settings_dict["nosetest_envar"] = "M" + args.obj.upper()[0] + '_HOST'

    # store cwd
    cwd = os.getcwd()

    # make command line args into dict
    clargs = dict([i.split('=', maxsplit=1) for i in args.o])

    # override any key value pairs from the command line
    settings_dict.update(clargs)

    if args.verbose:
        print("Creating {} project template using {}".format(args.b, settings_dict))

    # Make top level directory
    pdir = os.path.abspath(args.path)
    if args.verbose:
        print("Creating directory structure at: {}".format(pdir))

    # Start to template files out
    os.chdir(template_dir)

    # Template files out
    for (index, (dirpath, dirnames, filenames)) in enumerate(list(os.walk('.'))):
        if dirpath == '.':
            thisdir = os.path.join(pdir, settings_dict['src_package'])
        else:
            thisdir = os.path.join(pdir, settings_dict['src_package'], dirpath)
        os.mkdir(thisdir)
        for fi in [f for f in filenames if f.endswith('-tpl')]:
            with open(os.path.join(thisdir, fi[:-4]), 'w') as outfile,\
                    open(os.path.join(os.path.abspath(dirpath), fi), 'r') as infile:
                outfile.write(Template(infile.read()).substitute(settings_dict))
    os.chdir(cwd)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="A tool to create a skeleton biothings project",
        epilog="Available settings:\n{}".format(list(settings_dict.keys()))
    )
    parser.add_argument('obj', help="biothing type (e.g. gene, variant, drug, etc).  Default settings are created from this")
    parser.add_argument('path', help="path where skeleton project should be placed")
    parser.add_argument('-o', default=[], nargs="*", help="override any of the default settings with key=value pairs, to see default settings list, use -l")
    parser.add_argument('-b', default='es', help="type of database used for this biothing project")
    parser.add_argument('-v', default=False, dest="verbose", action="store_true", help="Verbose")
    args = parser.parse_args()
    main(args)



import random
import logging

try:
    import redis
except ImportError:
    logging.error('"redis" module is required to access Redis server.')


class RedisClientError(Exception):
    pass

class RedisClient(object):

    client = None

    @classmethod
    def get_client(cls, params):
        if cls.client is None:
            cls.client = cls(params)
        return cls.client

    def __init__(self, connection_params):
        self._mapdb = None
        self.connection_params = connection_params

    @property
    def mapdb(self):
        if self._mapdb is None:
            self._mapdb = redis.StrictRedis(db=0, **self.connection_params)
        return self._mapdb

    def get_db(self, db_name=None):
        """
        Return a redict client instance from a database name or
        database number (if db_name is an integer)
        """
        self.check()
        try:
            db_num = int(db_name)
        except ValueError:
            db_num = self.mapdb.get(db_name)
            if not db_num:
                db_num = self.pick_db()
        self.mapdb.set(db_name, int(db_num))

        return redis.StrictRedis(db=int(db_num), **self.connection_params)

    def pick_db(self):
        """
        Return a database number, preferably not used (db doesn't exist).
        If no database available (all are used), will be one and flush it...
        """
        db_max_num = int(self.mapdb.config_get("databases")["databases"] or 16)
        # -1: we always keep db=0 (meta db)
        avail = dict(zip(range(1, db_max_num), [True]*(db_max_num-1)))
        for info in self.mapdb.info("keyspace"):
            if not info.startswith("db"):
                continue
            num = int(info.replace("db", ""))
            if num == 0:
                continue
            avail.pop(num)
        if not avail:
            db_num = random.randint(1, db_max_num-1)
        else:
            db_num = random.choice(list(avail.keys()))

        return db_num

    def check(self):
        if not self.mapdb.get("__META__") == b'0':
            raise RedisClientError("Can't find database metadata, you may want to use initialize()")

    def initialize(self, deep=False):
        """
        Careful: this may delete data.
        Prepare Redis instance to work with biothings hub:
        - database 0: this db is used to store a mapping between
          database index and database name (so a database can be accessed
          by name). This method will flush this db and prepare it.
        - any other databases will be flushed if deep is True, making the redis
          server fully dedicated to
        """
        if deep:
            self.mapdb.flushall()
        self.mapdb.flushdb()
        self.mapdb.set("__META__", 0)
        self.check()



"""
Utility functions for parsing flatfiles,
mapping to JSON, cleaning.
"""
# see tabfile_feeder(coerce_unicode) if needed
#from __future__ import unicode_literals
import itertools
import csv
import os
import os.path
import json
import collections
from functools import total_ordering

from .common import open_anyfile, is_str, safewfile, anyfile
from .dotstring import key_value, set_key_value

csv.field_size_limit(10000000)   # default is 131072, too small for some big files


def dict_sweep(d, vals=None, remove_invalid_list=False):
    """
    Remove keys whos values are ".", "-", "", "NA", "none", " "
    and remove empty dictionaries

    Args:
        d (dict): a dictionary
        vals (str or list): a string or list of strings to sweep, or None to use the default values
        remove_invalid_list (boolean): when true, will remove key for which
            list has only one value, which is part of "vals".
            Ex::

                test_dict = {'gene': [None, None], 'site': ["Intron", None], 'snp_build' : 136}

            with `remove_invalid_list == False`::

                {'gene': [None], 'site': ['Intron'], 'snp_build': 136}

            with `remove_invalid_list == True`::

                {'site': ['Intron'], 'snp_build': 136}
    """
    # set default supported vals for empty values
    vals = vals or [".", "-", "", "NA", "none", " ", "Not Available", "unknown"]
    for key, val in list(d.items()):
        if val in vals:
            del d[key]
        elif isinstance(val, list):
            if remove_invalid_list:
                val = [v for v in val if v not in vals]
                for item in val:
                    if isinstance(item, dict):
                        dict_sweep(item, vals, remove_invalid_list=remove_invalid_list)
                # if len(val) == 0:
                if not val:
                    del d[key]
                else:
                    d[key] = val
            else:
                for item in val:
                    if item in vals:
                        val.remove(item)
                    elif isinstance(item, dict):
                        dict_sweep(item, vals, remove_invalid_list=remove_invalid_list)
                # if len(val) == 0:
                if not val:
                    del d[key]
        elif isinstance(val, dict):
            dict_sweep(val, vals, remove_invalid_list=remove_invalid_list)
            # if len(val) == 0:
            if not val:
                del d[key]
    return d


def safe_type(f, val):
    """
    Convert an input string to int/float/... using passed function.
    If the conversion fails then None is returned.
    If value of a type other than a string
    then the original value is returned.
    """
    if is_str(val):
        try:
            return f(val)
        except ValueError:
            pass
    return val


def to_float(val):
    """convert an input string to int"""
    return safe_type(float, val)

def to_int(val):
    """convert an input string to float"""
    return safe_type(int, val)

def to_number(val):
    """convert an input string to int/float."""
    if is_str(val):
        try:
            return int(val)
        except ValueError:
            try:
                return float(val)
            except ValueError:
                pass
    return val


def boolean_convert(d, convert_keys=None, level=0):
    """Explore document d and specified convert keys to boolean.
    Use dotfield notation for inner keys"""
    convert_keys = convert_keys or []
    for key, val in d.items():
        if isinstance(val, dict):
            d[key] = boolean_convert(val, convert_keys)
        if key in [ak.split(".")[level] for ak in convert_keys if len(ak.split(".")) > level]:
            if isinstance(val, list) or isinstance(val, tuple):
                if val and isinstance(val[0], dict):
                    d[key] = [boolean_convert(v, convert_keys, level+1) for v in val]
                else:
                    d[key] = [to_boolean(x) for x in val]
            elif isinstance(val, dict) or isinstance(val, collections.OrderedDict):
                d[key] = boolean_convert(val, convert_keys, level+1)
            else:
                d[key] = to_boolean(val)
    return d

def float_convert(d, include_keys=None, exclude_keys=None):
    """Convert elements in a document to floats.

    By default, traverse all keys
    If include_keys is specified, only convert the list from include_keys a.b, a.b.c
    If exclude_keys is specified, only exclude the list from exclude_keys

    :param d: a dictionary to traverse keys on
    :param include_keys: only convert these keys (optional)
    :param exclude_keys: exclude all other keys except these keys (optional)
    :return: generate key, value pairs
    """
    return value_convert_incexcl(d, to_float, include_keys, exclude_keys)

def int_convert(d, include_keys=None, exclude_keys=None):
    """Convert elements in a document to integers.

    By default, traverse all keys
    If include_keys is specified, only convert the list from include_keys a.b, a.b.c
    If exclude_keys is specified, only exclude the list from exclude_keys

    :param d: a dictionary to traverse keys on
    :param include_keys: only convert these keys (optional)
    :param exclude_keys: exclude all other keys except these keys (optional)
    :return: generate key, value pairs
    """
    return value_convert_incexcl(d, to_int, include_keys, exclude_keys)

def to_boolean(val, true_str=None, false_str=None):
    """Normlize str value to boolean value"""
    # set default true_str and false_str
    true_str = true_str or ['true', '1', 't', 'y', 'yes', 'Y', 'Yes', 'YES', 1]
    false_str = false_str or ['false', '0', 'f', 'n', 'N', 'No', 'no', 'NO', 0]
    # if type(val)!=str:
    if not isinstance(val, str):
        return bool(val)
    else:
        if val in true_str:
            return True
        elif val in false_str:
            return False


def merge_duplicate_rows(rows, db):
    """
    @param rows: rows to be grouped by
    @param db: database name, string
    """
    rows = list(rows)

    keys = set()
    for row in rows:
        for k in row[db]:
            keys.add(k)

    first_row = rows[0]
    other_rows = rows[1:]
    for row in other_rows:
        for i in keys:
            try:
                aa = first_row[db][i]
            except KeyError:
                try:
                    first_row[db][i] = row[db][i]
                except KeyError:
                    pass
                continue
            if i in row[db]:
                if row[db][i] != first_row[db][i]:
                    if not isinstance(aa, list):
                        aa = [aa]
                    aa.append(row[db][i])
                    first_row[db][i] = aa
            else:
                continue
    return first_row


def unique_ids(src_module):
    i = src_module.load_data()
    out = list(i)
    id_list = [a['_id'] for a in out if a]
    myset = set(id_list)
    print(len(out), "Documents produced")
    print(len(myset), "Unique IDs")
    return out


def rec_handler(infile, block_end='\n', skip=0, include_block_end=False, as_list=False):
    '''A generator to return a record (block of text)
       at once from the infile. The record is separated by
       one or more empty lines by default.
       skip can be used to skip top n-th lines
       if include_block_end is True, the line matching block_end will also be returned.
       if as_list is True, return a list of lines in one record.
    '''
    rec_separator = lambda line: line == block_end
    with open_anyfile(infile) as in_f:
        if skip:
            for i in range(skip):
                in_f.readline()
                del i
        for key, group in itertools.groupby(in_f, rec_separator):
            if not key:
                if include_block_end:
                    _g = itertools.chain(group, (block_end,))
                yield list(_g) if as_list else ''.join(_g)


#===============================================================================
# List Utility functions
#===============================================================================

# if dict value is a list of length 1, unlist
def unlist(d):
    for key, val in d.items():
        if isinstance(val, list):
            if len(val) == 1:
                d[key] = val[0]
        elif isinstance(val, dict):
            unlist(val)
    return d


def unlist_incexcl(d, include_keys=None, exclude_keys=None):
    """Unlist elements in a document.

    If there is 1 value in the list, set the element to that value.  Otherwise,
    leave the list unchanged.

    By default, traverse all keys
    If include_keys is specified, only traverse the list from include_keys a.b, a.b.c
    If exclude_keys is specified, only exclude the list from exclude_keys

    :param d: a dictionary to unlist
    :param include_keys: only unlist these keys (optional)
    :param exclude_keys: exclude all other keys except these keys (optional)
    :return: generate key, value pairs
    """
    def unlist_helper(d, include_keys=None, exclude_keys=None, keys=None):
        include_keys = include_keys or []
        exclude_keys = exclude_keys or []
        keys = keys or []
        if isinstance(d, dict):
            for key, val in d.items():
                if isinstance(val, list):
                    if len(val) == 1:
                        path = '.'.join(keys + [key])
                        if include_keys:
                            if path in include_keys:
                                d[key] = val[0]
                        elif path not in exclude_keys:
                            d[key] = val[0]
                elif isinstance(val, dict):
                    unlist_helper(val, include_keys, exclude_keys, keys + [key])
    unlist_helper(d, include_keys, exclude_keys, [])
    return d


def list_split(d, sep):
    """Split fields by sep into comma separated lists, strip."""
    for key, val in d.items():
        if isinstance(val, dict):
            list_split(val, sep)
        try:
            if len(val.split(sep)) > 1:
                d[key] = val.rstrip().rstrip(sep).split(sep)
        except AttributeError:
            pass
    return d


def id_strip(id_list):
    id_list = id_list.split("|")
    ids = []
    for _id in id_list:
        ids.append(_id.rstrip().lstrip())
    return ids

def llist(li, sep='\t'):
    '''Nicely output the list with each item a line.'''
    for x in li:
        if isinstance(x, (li, tuple)):
            xx = sep.join([str(i) for i in x])
        else:
            xx = str(x)
        print(xx)


def listitems(a_list, *idx):
    '''Return multiple items from list by given indexes.'''
    if isinstance(a_list, tuple):
        # return tuple([a_list[i] for i in idx])   # TODO: remove this line
        return tuple(a_list[i] for i in idx)
    else:
        return [a_list[i] for i in idx]


def list2dict(a_list, keyitem, alwayslist=False):     # pylint: disable=redefined-outer-name
    '''Return a dictionary with specified keyitem as key, others as values.
       keyitem can be an index or a sequence of indexes.
       For example::

            li=[['A','a',1],
                ['B','a',2],
                ['A','b',3]]
            list2dict(li,0)---> {'A':[('a',1),('b',3)],
                                 'B':('a',2)}

       If `alwayslist` is True, values are always a list even there is only one item in it::

            list2dict(li,0,True)---> {'A':[('a',1),('b',3)],
                                      'B':[('a',2),]}
    '''
    _dict = {}
    for x in a_list:
        if isinstance(keyitem, int):      # single item as key
            key = x[keyitem]
            value = tuple(x[:keyitem] + x[keyitem + 1:])
        else:
            # key = tuple([x[i] for i in keyitem])  # TODO: remove this line
            key = tuple(x[i] for i in keyitem)
            # value = tuple([x[i] for i in range(len(a_list)) if i not in keyitem])  # TODO: remove this line
            value = tuple(x[i] for i in range(len(a_list)) if i not in keyitem)
        if len(value) == 1:      # single value
            value = value[0]
        if key not in _dict:
            if alwayslist:
                _dict[key] = [value, ]
            else:
                _dict[key] = value
        else:
            current_value = _dict[key]
            if not isinstance(current_value, list):
                current_value = [current_value, ]
            current_value.append(value)
            _dict[key] = current_value
    return _dict


def list_nondup(a_list):
    # TODO: use set in caller and remove func
    x = {}
    for item in a_list:
        x[item] = None
    return x.keys()


def listsort(a_list, by, reverse=False, cmp=None, key=None):
    '''Given list is a list of sub(list/tuple.)
       Return a new list sorted by the ith(given from "by" item)
       item of each sublist.'''
    new_li = [(x[by], x) for x in a_list]
    new_li.sort(cmp=cmp, key=key, reverse=reverse)
    return [x[1] for x in new_li]


def list_itemcnt(a_list):
    '''Return number of occurrence for each type of item in the list.'''
    x = {}
    for item in a_list:
        # TODO= use list.count(value)
        if item in x:
            x[item] += 1
        else:
            x[item] = 1
    return [(i, x[i]) for i in x]


def alwayslist(value):
    """If input value if not a list/tuple type, return it as a single value list."""
    if value is None:
        return []
    if isinstance(value, (list, tuple)):
        return value
    else:
        return [value]

#===============================================================================
# File Utility functions
#===============================================================================

def tabfile_tester(datafile, header=1, sep='\t'):
    reader = csv.reader(anyfile(datafile), delimiter=sep)
    lineno = 0
    try:
        for i in range(header):
            del i
            next(reader)
            lineno += 1

        for ld in reader:
            del ld
            lineno += 1
    except:
        print("Error at line number:", lineno)
        raise


def dupline_seperator(dupline, dup_sep, dup_idx=None, strip=False):
    '''
    for a line like this::

        a   b1,b2  c1,c2

    return a generator of this list (breaking out of the duplicates in each field)::

        [(a,b1,c1),
         (a,b2,c1),
         (a,b1,c2),
         (a,b2,c2)]

    Example::

         dupline_seperator(dupline=['a', 'b1,b2', 'c1,c2'],
                           dup_idx=[1,2],
                           dup_sep=',')

    if dup_idx is None, try to split on every field.
    if strip is True, also tripe out of extra spaces.
    '''
    value_li = list(dupline)
    for idx, value in enumerate(value_li):
        if dup_idx:
            if idx in dup_idx:
                value = value.split(dup_sep)
                if strip:
                    value = [x.strip() for x in value]
            else:
                value = [value]
        else:
            value = value.split(dup_sep)
            if strip:
                value = [x.strip() for x in value]
        value_li[idx] = value
    return itertools.product(*value_li)    # itertools.product fits exactly the purpose here


def tabfile_feeder(datafile, header=1, sep='\t',
                   includefn=None,
                   coerce_unicode=True,
                   assert_column_no=None):
    '''a generator for each row in the file.'''

    in_f = anyfile(datafile)
    reader = csv.reader(in_f, delimiter=sep)
    lineno = 0
    try:
        for i in range(header):
            next(reader)
            lineno += 1
            del i

        for ld in reader:
            if assert_column_no:
                if len(ld) != assert_column_no:
                    err = "Unexpected column number:" \
                          " got {}, should be {}".format(len(ld), assert_column_no)
                    raise ValueError(err)
            if not includefn or includefn(ld):
                lineno += 1
                if coerce_unicode:
                    yield [str(x) for x in ld]
                else:
                    yield ld
    except ValueError:
        print("Error at line number:", lineno)
        raise


def tab2list(datafile, cols, **kwargs):
    if os.path.exists(datafile):
        if isinstance(cols, int):
            return [ld[cols] for ld in tabfile_feeder(datafile, **kwargs)]
        else:
            return [listitems(ld, *cols) for ld in tabfile_feeder(datafile, **kwargs)]
    else:
        print('Error: missing "%s". Skipped!' % os.path.split(datafile)[1])
        return {}


def tab2dict(datafile, cols, key, alwayslist=False, **kwargs):     # pylint: disable=redefined-outer-name
    if isinstance(datafile, tuple):
        _datafile = datafile[0]
    else:
        _datafile = datafile
    if os.path.exists(_datafile):
        return list2dict([listitems(ld, *cols) for ld in tabfile_feeder(datafile, **kwargs)], key, alwayslist=alwayslist)
    else:
        print('Error: missing "%s". Skipped!' % os.path.split(_datafile)[1])
        return {}


def tab2dict_iter(datafile, cols, key, alwayslist=False, **kwargs):     # pylint: disable=redefined-outer-name
    if isinstance(datafile, tuple):
        _datafile = datafile[0]
    else:
        _datafile = datafile
    if os.path.exists(_datafile):
        bulk = []
        prev_id = None
        for ld in tabfile_feeder(datafile, **kwargs):
            #print(ld)
            li = listitems(ld, *cols)
            #print("key %s len bulk %s prev %s" % (li[key],len(bulk),prev_id))
            if prev_id is None or (li[key] == prev_id):
                #print("\t\tfound same")
                bulk.append(li)
                prev_id = li[key]
            else:
                #print("bulk size: %s" % bulk)
                di = list2dict(bulk, key, alwayslist=alwayslist)
                bulk = []
                # changed key, init next bulk
                bulk.append(li)
                prev_id = li[key]
                #print("on yield: %s" % di)
                yield di
        # flush remaining bulk
        if bulk:
            di = list2dict(bulk, key, alwayslist=alwayslist)
            yield di
    else:
        print('Error: missing "%s". Skipped!' % os.path.split(_datafile)[1])
        return {}


def file_merge(infiles, outfile=None, header=1, verbose=1):
    '''merge a list of input files with the same format.
       if header will be removed from the 2nd files in the list.
    '''
    outfile = outfile or '_merged'.join(os.path.splitext(infiles[0]))
    out_f, outfile = safewfile(outfile)
    if verbose:
        print("Merging...")
    cnt = 0
    for i, fn in enumerate(infiles):
        print(os.path.split(fn)[1], '...', end='')
        line_no = 0
        in_f = anyfile(fn)
        if i > 0:
            for k in range(header):
                in_f.readline()
                del k
        for line in in_f:
            out_f.write(line)
            line_no += 1
        in_f.close()
        cnt += line_no
        print(line_no)
    out_f.close()
    print("=" * 20)
    print("Done![total %d lines output]" % cnt)

#===============================================================================
# Dictionary & other structures Utility functions
#===============================================================================

# http://stackoverflow.com/questions/12971631/sorting-list-by-an-attribute-that-can-be-none
# used to sort list with None element (because python3 suddenly decided it wwasn't possible
# anymore. because...)
# from functools import total_ordering
@total_ordering
class MinType(object):
    def __le__(self, other):
        return True

    def __eq__(self, other):
        return self is other
Min = MinType()


def traverse_keys(d, include_keys=None, exclude_keys=None):
    """Return all key, value pairs for a document.

    By default, traverse all keys
    If include_keys is specified, only traverse the list from include_kes a.b, a.b.c
    If exclude_keys is specified, only exclude the list from exclude_keys

    :param d: a dictionary to traverse keys on
    :param include_keys: only traverse these keys (optional)
    :param exclude_keys: exclude all other keys except these keys (optional)
    :return: generate key, value pairs
    """
    include_keys = include_keys or []
    exclude_keys = exclude_keys or []

    def traverse_helper(d, keys):
        if isinstance(d, dict):
            for k in d.keys():
                yield from traverse_helper(d[k], keys + [k])
        elif isinstance(d, list):
            for i in d:
                yield from traverse_helper(i, keys)
        else:
            yield keys, d

    if include_keys:
        for k in include_keys:
            for val in key_value(d, k):
                yield k, val
    else:
        for kl, val in traverse_helper(d, []):
            key = '.'.join(kl)
            if key not in exclude_keys:
                yield key, val


# from mygene, originally
def value_convert(_dict, fn, traverse_list=True):
    '''For each value in _dict, apply fn and then update
       _dict with return the value.
       if traverse_list is True and a value is a list,
       apply fn to each item of the list.
    '''
    for k in _dict:
        if traverse_list and isinstance(_dict[k], list):
            _dict[k] = [fn(x) for x in _dict[k]]
        else:
            _dict[k] = fn(_dict[k])
    return _dict


def value_convert_incexcl(d, fn, include_keys=None, exclude_keys=None):
    """Convert elements in a document using a function fn.

    By default, traverse all keys
    If include_keys is specified, only convert the list from include_keys a.b, a.b.c
    If exclude_keys is specified, only exclude the list from exclude_keys

    :param d: a dictionary to traverse keys on
    :param fn: function to convert elements with
    :param include_keys: only convert these keys (optional)
    :param exclude_keys: exclude all other keys except these keys (optional)
    :return: generate key, value pairs
    """
    for path, value in traverse_keys(d, include_keys, exclude_keys):
        new_value = fn(value)
        set_key_value(d, path, new_value)
    return d


# from biothings, originally
# closed to value_convert, could be refactored except this one
# is recursive for dict typed values
def value_convert_to_number(d, skipped_keys=None):
    """convert string numbers into integers or floats
       skip converting certain keys in skipped_keys list"""
    skipped_keys = skipped_keys or []
    for key, val in d.items():
        if isinstance(val, dict):
            value_convert_to_number(val, skipped_keys)
        if key not in skipped_keys:
            if isinstance(val, list):
                d[key] = [to_number(x) if not isinstance(x, dict) else value_convert_to_number(x, skipped_keys) for x in val]
            elif isinstance(val, tuple):
                # TODO: remove this line
                # d[key] = tuple([to_number(x) if not isinstance(x, dict) else value_convert_to_number(x, skipped_keys) for x in val])
                d[key] = tuple(to_number(x) if not isinstance(x, dict) else value_convert_to_number(x, skipped_keys) for x in val)
            else:
                d[key] = to_number(val)
    return d

def dict_convert(_dict, keyfn=None, valuefn=None):
    '''Return a new dict with each key converted by keyfn (if not None),
       and each value converted by valuefn (if not None).
    '''
    if keyfn is None and valuefn is not None:
        for k in _dict:
            _dict[k] = valuefn(_dict[k])
        return _dict

    elif keyfn is not None:
        out_dict = {}
        for k in _dict:
            out_dict[keyfn(k)] = valuefn(_dict[k]) if valuefn else _dict[k]
        return out_dict
    else:
        return _dict

def updated_dict(_dict, attrs):
    '''Same as dict.update, but return the updated dictionary.'''
    out = _dict.copy()
    out.update(attrs)
    return out

def update_dict_recur(d, u):
    """
    Update dict d with dict u's values, recursively
    (so existing values in d but not in u are kept even if nested)
    """
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            r = update_dict_recur(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d

def merge_dict(dict_li, attr_li, missingvalue=None):
    '''
    Merging multiple dictionaries into a new one.
    Example::

        In [136]: d1 = {'id1': 100, 'id2': 200}
        In [137]: d2 = {'id1': 'aaa', 'id2': 'bbb', 'id3': 'ccc'}
        In [138]: merge_dict([d1,d2], ['number', 'string'])
        Out[138]:
        {'id1': {'number': 100, 'string': 'aaa'},
        'id2': {'number': 200, 'string': 'bbb'},
        'id3': {'string': 'ccc'}}
        In [139]: merge_dict([d1,d2], ['number', 'string'], missingvalue='NA')
        Out[139]:
        {'id1': {'number': 100, 'string': 'aaa'},
        'id2': {'number': 200, 'string': 'bbb'},
        'id3': {'number': 'NA', 'string': 'ccc'}}
    '''
    dd = dict(zip(attr_li, dict_li))
    key_set = set()
    for attr in dd:
        key_set = key_set | set(dd[attr])

    out_dict = {}
    for k in key_set:
        value = {}
        for attr in dd:
            if k in dd[attr]:
                value[attr] = dd[attr][k]
            elif missingvalue is not None:
                value[attr] = missingvalue
        out_dict[k] = value
    return out_dict

def normalized_value(value, sort=True):
    '''Return a "normalized" value:
           1. if a list, remove duplicate and sort it
           2. if a list with one item, convert to that single item only
           3. if a list, remove empty values
           4. otherwise, return value as it is.
    '''
    if isinstance(value, list):
        value = [x for x in value if x]   # remove empty values
        try:
            _v = list(set(value))
        except TypeError:
            #use alternative way
            # _v = [json.loads(x) for x in set([json.dumps(x) for x in value])]   # TODO: remove this line
            _v = [json.loads(x) for x in {json.dumps(x) for x in value}]
        # if len(_v) and sort:
        if _v and sort:
            # py3 won't sort dict anymore...
            if isinstance(_v[0], dict):
                _v = sorted(_v, key=lambda x: sorted(x.keys()))
            else:
                try:
                    _v = sorted(_v)
                except TypeError:
                    # probably some None values to sort, not handle anymore in py3
                    # let's use a trick...
                    _v = sorted(_v, key=lambda x: Min if x is None or (not isinstance(x, str) and None in x) else x)
        if len(_v) == 1:
            _v = _v[0]
    else:
        _v = value

    return _v

def dict_nodup(_dict, sort=True):
    for k in _dict:
        _dict[k] = normalized_value(_dict[k], sort=sort)
    return _dict

def dict_attrmerge(dict_li, removedup=True, sort=True, special_fns=None):
    '''
        dict_attrmerge([{'a': 1, 'b':[2,3]},
                        {'a': [1,2], 'b':[3,5], 'c'=4}])
        sould return
             {'a': [1,2], 'b':[2,3,5], 'c'=4}

        special_fns is a dictionary of {attr:  merge_fn}
         used for some special attr, which need special merge_fn
         e.g.,   {'uniprot': _merge_uniprot}
    '''
    special_fns = special_fns or {}
    out_dict = {}
    keys = []
    for d in dict_li:
        keys.extend(d.keys())
    keys = set(keys)
    for k in keys:
        _value = []
        for d in dict_li:
            if d.get(k, None):
                if isinstance(d[k], list):
                    _value.extend(d[k])
                else:
                    _value.append(d[k])
        if len(_value) == 1:
            out_dict[k] = _value[0]
        else:
            out_dict[k] = _value

        if k in special_fns:
            out_dict[k] = special_fns[k](out_dict[k])

    if removedup:
        out_dict = dict_nodup(out_dict, sort=sort)
    return out_dict

def merge_root_keys(doc1, doc2, exclude=None):
    """
    Ex: d1 = {"_id":1,"a":"a","b":{"k":"b"}}
        d2 = {"_id":1,"a":"A","b":{"k":"B"},"c":123}

        Both documents have the same _id, and 2 root keys, "a" and "b".
        Using this storage, the resulting document will be:

        {'_id': 1, 'a': ['A', 'a'], 'b': [{'k': 'B'}, {'k': 'b'}],"c":123}
    """
    # we'll "eat" from doc2 so clean it first as needed
    exclude = exclude or []
    for k in exclude:
        doc2.pop(k, None)
    for k1 in doc1:
        if k1 in exclude:
            continue
        v2 = doc2.pop(k1, None)
        if not isinstance(v2, list):
            v2 = [v2]
        if v2:
            if isinstance(doc1[k1], list):
                doc1[k1].extend(v2)
            else:
                doc1[k1] = [doc1[k1]] + v2
    # merge what's remaining in doc2 that wasn't in doc1
    doc1.update(doc2)

    return doc1

def dict_apply(d, key, value, sort=True):
    """add value to d[key], append it if key exists

        >>> d = {'a': 1}
        >>> dict_apply(d, 'a', 2)
         {'a': [1, 2]}
        >>> dict_apply(d, 'a', 3)
         {'a': [1, 2, 3]}
        >>> dict_apply(d, 'b', 2)
         {'a': 1, 'b': 2}
    """
    if key in d:
        _value = d[key]
        if not isinstance(_value, list):
            _value = [_value]
        if isinstance(value, list):
            _value.extend(value)
        else:
            _value.append(value)
    else:
        _value = value

    d[key] = normalized_value(_value, sort=sort)

def dict_to_list(gene_d):
    '''return a list of genedoc from genedoc dictionary and
       make sure the "_id" field exists.
    '''
    doc_li = [updated_dict(gene_d[k], {'_id': str(k)}) for k in sorted(gene_d.keys())]
    return doc_li

def merge_struct(v1, v2, aslistofdict=None):

    #print("v1 = %s" % repr(v1))
    #print("v2 = %s" % repr(v2))

    if isinstance(v1, list):
        #print("v1 is list ", end="")
        if isinstance(v2, list):
            #print("v2 is list -> extend")
            #v1.extend(v2)
            #v1 = list(set(v1))
            v1 = v1 + [x for x in v2 if x not in v1]
        else:
            #print("v2 not list -> append")
            if v2 not in v1:
                v1.append(v2)

    elif isinstance(v2, list) and isinstance(v1, dict):
        #return merge_struct(v2,v1)
        if v1 not in v2:
            v2.append(v1)

    elif isinstance(v1, dict):
        assert isinstance(v2, dict), "v2 %s not a dict (v1: %s)" % (v2, v1)
        #print("v1 & v2 is dict")
        for k in list(v1.keys()):
            #print("v1[%s]" % k)
            if k in v2:
                #print("%s is both dict -> merge/update v1[%s],v2[%s]" % (k, k, k))
                #v1.update(merge(v1[k], v2[k]))
                if aslistofdict == k:
                    v1elem = v1[k]
                    v2elem = v2[k]
                    if not isinstance(v1elem, list):
                        v1elem = [v1elem]
                    if not isinstance(v2elem, list):
                        v2elem = [v2elem]
                    # v1elem and v2elem may be the same, in this case as a result
                    # we may have transformed it in a list (no merge, but just type change).
                    # if so, back to scalar
                    if v1elem != v2elem:
                        v1[k] = merge_struct(v1elem, v2elem)
                else:
                    v1[k] = merge_struct(v1[k], v2[k])
            else:
                #print("%s not in v2 -> update v1 with v2" % k)
                #print("v2 before %s" % repr(v2))
                #v1.update(v2)
                v2[k] = v1[k]
                #print("v2 after %s" % repr(v2))
        for k in v2:
            #print("v2[%s]" % k)
            if k in v1:
                pass  # already done
            else:
                v1[k] = v2[k]

    elif isinstance(v1, str) or isinstance(v1, int) or isinstance(v1, float):
        if isinstance(v2, str) or isinstance(v2, int) or isinstance(v2, float):
            if v1 != v2:
                #print("v1 & v2 not iterable -> list of 2")
                v1 = [v1, v2]
            else:
                pass
                #print("v1 == v2, skip")
        else:
            #print("v2 iterable, reverse merge")
            return merge_struct(v2, v1)
    else:
        raise TypeError("dunno how to merge type %s" % type(v1))

    #print("return %s" % v1)
    return v1

def dict_walk(dictionary, key_func):
    """Recursively apply key_func to dict's keys"""
    if not isinstance(dictionary, dict):
        return dictionary
    # TODO: remove the following two lines
    # return dict((key_func(k), dict_walk(v, key_func))
    #             for k, v in dictionary.items())
    return {key_func(k): dict_walk(v, key_func) for k, v in dictionary.items()}


def dict_traverse(d, func, traverse_list=False):
    """
    Recursively traverse dictionary d, calling func(k,v)
    for each key/value found. func must return a
    tuple(new_key,new_value)
    """
    try:
        items = sorted(d.items(), key=lambda x: x[0])
    except TypeError:
        # not sortable
        items = d.items()
    for k, v in items:
        if isinstance(v, dict):
            dict_traverse(v, func, traverse_list=traverse_list)
        elif traverse_list and isinstance(v, list):
            for e in v:
                dict_traverse(e, func, traverse_list=traverse_list)
        else:
            newk, newv = func(k, v)
            d.pop(k)
            d[newk] = newv



KWARG_DESCRIPTIONS = {
    '_source': {
        'name': 'fields',
        'text_template': 'a comma-separated list of fields (in dotfield notation) used to limit the fields returned from the matching {biothing_object} hit(s). The supported field names can be found from any {biothing_object} object or from the /metadata/fields endpoint. If "fields=all", all available fields will be returned.{param_type}{param_default_value}{param_max}'},
    'size': {
        'name': 'size',
        'text_template': 'the maximum number of matching {biothing_object} hits to return per batch.{param_type}{param_default_value}{param_max}'},
    'from': {
        'name': 'from',
        'text_template': 'the number of matching {biothing_object} hits to skip, starting from 0.  This can be useful for paging in combination with the "size" parameter.{param_type}{param_default_value}{param_max}'},
    'sort': {
        'name': 'sort',
        'text_template': 'the comma-separated list of fields to sort on. Prefix each with "-" for descending order, otherwise in ascending order. Default: sort by descending score.'},
    'dotfield': {
        'name': 'dotfield',
        'text_template': 'control the format of the returned {biothing_object} object. If "true" or "1", all fields will be collapsed into a single level deep object (all nested objects will be a single level deep, using dotfield notation to signify the nested structure){param_type}{param_default_value}{param_max}'},
    'email': {
        'name': 'email',
        'text_template': 'If you are regular users of our services, we encourage you to provide us with an email, so that we can better track the usage or follow up with you.'},
    'out_format': {
        'name': 'format',
        'text_template': 'controls output format of server response, currently supports: "json", "jsonld", "html".{param_type}{param_default_value}{param_max}'},
    'aggs': {
        'name': 'facets',
        'text_template': 'a comma-separated list of fields to return facets on.  In addition to query hits, the fields notated in "facets" will be aggregated by value and bucklet counts will be displayed in the "facets" field of the response object.{param_type}{param_default_value}{param_max}'},
    'facet_size': {
        'name': 'facet_size',
        'text_template': 'the number of facet buckets to return in the response.{param_type}{param_default_value}{param_max}'},
    'ids': {
        'name': 'ids',
        'text_template': 'multiple {biothing_object} ids separated by comma. Note that currently we only take the input ids up to 1000 maximum, the rest will be omitted.{param_type}{param_default_value}{param_max}'},
    'q': {
        'name': 'q',
        'text_template': 'Query string.  The detailed query syntax can be found from our [docs]{doc_query_syntax_url}'},
    'scopes': {
        'name': 'scopes',
        'text_template': 'a comma-separated list of fields as the search "scopes" (fields to search through for query term). The available "fields" that can be passed to the "scopes" parameter are listed in the **/metadata/fields** endpoint.{param_type} Default: "scopes=_id".{param_max}'},
    'search': {
        'name': 'search',
        'text_template': 'Pass a search term to filter the available fields.{param_type}{param_default_value}{param_max}'},
    'prefix': {
        'name': 'prefix',
        'text_template': 'Pass a prefix string to filter the available fields.{param_type}{param_default_value}{param_max}'}}



'''
Utils to compare two list of gene documents
'''
import os
import os.path
import time
from .common import timesofar, dump, get_timestamp, filter_dict
from .backend import DocMongoDBBackend
from ..hub.databuild.backend import create_backend
from .es import ESIndexer
from .jsondiff import make as jsondiff


def diff_doc(doc_1, doc_2, exclude_attrs=['_timestamp']):
    diff_d = {'update': {},
              'delete': [],
              'add': {}}
    if exclude_attrs:
        doc_1 = filter_dict(doc_1, exclude_attrs)
        doc_2 = filter_dict(doc_2, exclude_attrs)
    for attr in set(doc_1) | set(doc_2):
        if exclude_attrs and attr in exclude_attrs:
            continue
        if attr in doc_1 and attr in doc_2:
            _v1 = doc_1[attr]
            _v2 = doc_2[attr]
            if _v1 != _v2:
                diff_d['update'][attr] = _v2
        elif attr in doc_1 and attr not in doc_2:
            diff_d['delete'].append(attr)
        else:
            diff_d['add'][attr] = doc_2[attr]
    if diff_d['update'] or diff_d['delete'] or diff_d['add']:
        return diff_d


def full_diff_doc(doc_1, doc_2, exclude_attrs=['_timestamp']):
    diff_d = {'update': {},
              'delete': [],
              'add': {}}
    if exclude_attrs:
        doc_1 = filter_dict(doc_1, exclude_attrs)
        doc_2 = filter_dict(doc_2, exclude_attrs)
    for attr in set(doc_1) | set(doc_2):
        if exclude_attrs and attr in exclude_attrs:
            continue
        if attr in doc_1 and attr in doc_2:
            _v1 = doc_1[attr]
            _v2 = doc_2[attr]
            difffound = False
            if isinstance(_v1, dict) and isinstance(_v2, dict):
                if full_diff_doc(_v1, _v2, exclude_attrs):
                    difffound = True
            elif isinstance(_v1, list) and isinstance(_v2, list):
                # there can be unhashable/unordered dict in these lists
                for i in _v1:
                    if i not in _v2:
                        difffound = True
                        break
                # check the other way
                if not difffound:
                    for i in _v2:
                        if i not in _v1:
                            difffound = True
                            break
            elif _v1 != _v2:
                difffound = True

            if difffound:
                diff_d['update'][attr] = _v2

        elif attr in doc_1 and attr not in doc_2:
            diff_d['delete'].append(attr)
        else:
            diff_d['add'][attr] = doc_2[attr]
    if diff_d['update'] or diff_d['delete'] or diff_d['add']:
        return diff_d

def two_docs_iterator(b1, b2, id_list, step=10000, verbose=False):
    t0 = time.time()
    n = len(id_list)
    for i in range(0, n, step):
        t1 = time.time()
        if verbose:
            print("Processing %d-%d documents..." % (i + 1, min(i + step, n)), end='')
        _ids = id_list[i:i+step]
        iter1 = sorted([d for d in b1.mget_from_ids(_ids, asiter=True)],key=lambda a: a["_id"])
        iter2 = sorted([d for d in b2.mget_from_ids(_ids, asiter=True)],key=lambda a: a["_id"])
        for doc1, doc2 in zip(iter1, iter2):
            yield doc1, doc2
        if verbose:
            print('Done.[%.1f%%,%s]' % (i*100./n, timesofar(t1)))
    if verbose:
        print("="*20)
        print('Finished.[total time: %s]' % timesofar(t0))


def _diff_doc_worker(args):
    _b1, _b2, ids, _path = args
    import biothings.utils.diff
    import importlib
    importlib.reload(biothings.utils.diff)
    from biothings.utils.diff import _diff_doc_inner_worker, get_backend

    b1 = get_backend(*_b1)
    b2 = get_backend(*_b2)

    _updates = _diff_doc_inner_worker(b1, b2, ids)
    return _updates


def _diff_doc_inner_worker(b1, b2, ids, fastdiff=False, diff_func=full_diff_doc):
    '''if fastdiff is True, only compare the whole doc,
       do not traverse into each attributes.
    '''
    _updates = []
    for doc1, doc2 in two_docs_iterator(b1, b2, ids):
        assert doc1['_id'] == doc2['_id'], repr((ids, len(ids)))
        if fastdiff:
            if doc1 != doc2:
                _updates.append({'_id': doc1['_id']})
        else:
            _diff = diff_func(doc1, doc2)
            if _diff:
                _diff['_id'] = doc1['_id']
                _updates.append(_diff)
    return _updates

def diff_docs_jsonpatch(b1, b2, ids, fastdiff=False, exclude_attrs=[]):
    '''if fastdiff is True, only compare the whole doc,
       do not traverse into each attributes.
    '''
    _updates = []
    for doc1, doc2 in two_docs_iterator(b1, b2, ids):
        assert doc1['_id'] == doc2['_id'], "Different ids: '%s' != '%s'" % \
                (doc1['_id'], doc2['_id'])
        if exclude_attrs:
            doc1 = filter_dict(doc1, exclude_attrs)
            doc2 = filter_dict(doc2, exclude_attrs)
        if fastdiff:
            if doc1 != doc2:
                _updates.append(doc1['_id'])
        else:
            _patch = jsondiff(doc1, doc2)
            if _patch:
                _diff = {}
                _diff['patch'] = _patch
                _diff['_id'] = doc1['_id']
                _updates.append(_diff)
    return _updates




# TODO: move to mongodb backend class
def get_mongodb_uri(backend):
    opt = backend.target_collection.database.client._MongoClient__options.credentials
    username = opt and opt.username or None
    password = opt and opt.password or None
    dbase = opt and opt.source or None
    uri = "mongodb://"
    if username:
        if password:
            uri += "%s:%s@" % (username, password)
        else:
            uri += "%s@" % username
    host, port = backend.target_collection.database.client.address
    uri += "%s:%s" % (host, port)
    uri += "/%s" % (dbase or backend.target_collection.database.name)
    #uri += "/%s" % backend.target_collection.name
    print("uri: %s" % uri)
    return uri


def diff_collections(b1, b2, use_parallel=True, step=10000):
    """
    b1, b2 are one of supported backend class in databuild.backend.
    e.g.::

        b1 = DocMongoDBBackend(c1)
        b2 = DocMongoDBBackend(c2)
    """

    id_s1 = set(b1.get_id_list())
    id_s2 = set(b2.get_id_list())
    print("Size of collection 1:\t", len(id_s1))
    print("Size of collection 2:\t", len(id_s2))

    id_in_1 = id_s1 - id_s2
    id_in_2 = id_s2 - id_s1
    id_common = id_s1 & id_s2
    print("# of docs found only in collection 1:\t", len(id_in_1))
    print("# of docs found only in collection 2:\t", len(id_in_2))
    print("# of docs found in both collections:\t", len(id_common))

    print("Comparing matching docs...")
    _updates = []
    if len(id_common) > 0:
        if not use_parallel:
            _updates = _diff_doc_inner_worker(b1, b2, list(id_common))
        else:
            from .parallel import run_jobs_on_ipythoncluster
            _path = os.path.split(os.path.split(os.path.abspath(__file__))[0])[0] + "/.."
            id_common = list(id_common)
            _b1 = (get_mongodb_uri(b1), b1.target_collection.database.name, b1.target_name, b1.name)
            _b2 = (get_mongodb_uri(b2), b2.target_collection.database.name, b2.target_name, b2.name)
            task_li = [(_b1, _b2, id_common[i: i + step], _path) for i in range(0, len(id_common), step)]
            job_results = run_jobs_on_ipythoncluster(_diff_doc_worker, task_li)
            _updates = []
            if job_results:
                for res in job_results:
                    _updates.extend(res)
            else:
                print("Parallel jobs failed or were interrupted.")
                return None

        print("Done. [{} docs changed]".format(len(_updates)))

    _deletes = []
    if len(id_in_1) > 0:
        _deletes = sorted(id_in_1)

    _adds = []
    if len(id_in_2) > 0:
        _adds = sorted(id_in_2)

    changes = {'update': _updates,
               'delete': _deletes,
               'add': _adds}
    return changes


def get_backend(uri, db, col, bk_type):
    if bk_type != "mongodb":
        raise NotImplementedError("Backend type '%s' not supported" % bk_type)
    from biothings.utils.mongo import MongoClient
    colobj = MongoClient(uri)[db][col]
    return DocMongoDBBackend(colobj)


def diff_collections_batches(b1, b2, result_dir, step=10000):
    '''
    b2 is new collection, b1 is old collection
    '''
    DIFFFILE_PATH = '/home/kevinxin/diff_result/'
    DATA_FOLDER = os.path.join(DIFFFILE_PATH, result_dir)
    if not os.path.exists(DATA_FOLDER):
        os.mkdir(DATA_FOLDER)
    data_new = doc_feeder(b2.target_collection, step=step, inbatch=True, fields=[])
    data_old = doc_feeder(b1.target_collection, step=step, inbatch=True, fields=[])
    cnt = 0
    cnt_update = 0
    cnt_add = 0
    cnt_delete = 0

    for _batch in data_new:
        cnt += 1
        id_list_new = [_doc['_id'] for _doc in _batch]
        docs_common = b1.target_collection.find({'_id': {'$in': id_list_new}}, projection=[])
        ids_common = [_doc['_id'] for _doc in docs_common]
        id_in_new = list(set(id_list_new) - set(ids_common))
        _updates = []
        if len(ids_common) > 0:
            _updates = diff_docs_jsonpatch(b1, b2, list(ids_common), fastdiff=True)
        file_name = DATA_FOLDER + '/' + str(cnt) + '.pyobj'
        _result = {'add': id_in_new,
                   'update': _updates,
                   'delete': [],
                   'source': b2.target_collection.name,
                   'timestamp': get_timestamp()}
        if len(_updates) != 0 or len(id_in_new) != 0:
            dump(_result, file_name)
            print("(Updated: {}, Added: {})".format(len(_updates), len(id_in_new)), end='')
            cnt_update += len(_updates)
            cnt_add += len(id_in_new)
    print("Finished calculating diff for the new collection. Total number of docs updated: {}, added: {}".format(cnt_update, cnt_add))
    print("="*100)
    for _batch in data_old:
        cnt += 1
        id_list_old = [_doc['_id'] for _doc in _batch]
        docs_common = b2.target_collection.find({'_id': {'$in': id_list_old}}, projection=[])
        ids_common = [_doc['_id'] for _doc in docs_common]
        id_in_old = list(set(id_list_old)-set(ids_common))
        file_name = DATA_FOLDER + '/' + str(cnt) + '.pyobj'
        _result = {'delete': id_in_old,
                   'add': [],
                   'update': [],
                   'source': b2.target_collection.name,
                   'timestamp': get_timestamp()}
        if len(id_in_old) != 0:
            dump(_result, file_name)
            print("(Deleted: {})".format(len(id_in_old)), end='')
            cnt_delete += len(id_in_old)
    print("Finished calculating diff for the old collection. Total number of docs deleted: {}".format(cnt_delete))
    print("="*100)
    print("Summary: (Updated: {}, Added: {}, Deleted: {})".format(cnt_update, cnt_add, cnt_delete))
